#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
  Execute a (list of) commands taking arguments from a file.
  Usage:
      dirac-loop [File] Command [Command2...] [option|cfgfile]

  Arguments:
     File: File containing the strings to be appended to the command (default /dev/stdin)
           If the arguments are BK paths, the script reduces the paths merging event and file types
           (use --NoMerge to disable)
     Command: Command to execute (separate with spaces)
              If command terminates with ", a " is appended after the argument
              If the command contains the string @arg@,
                it is replaced by the argument rather than the argument appended
              If the command itself is dirac-loop, then the first occurrence only of @arg@ is replaced,
              which allows recursive calls
"""
import os
import tempfile
import subprocess
from collections import defaultdict


from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.List import breakListIntoChunks


def reduceArgs(noMerge, arguments, maxList=20):
    """
    If the arguments look like BK paths (start with /LHCb or /MC), try to
    reduce the list of BK paths by merging event types or file types into a
    list.

    :param bool: if set, do not merge arguments if BK paths
    :param list arguments: list of arguments
    :param int maxList: maximum number of event/file types to put in a resulting merged argument
    """
    if noMerge:
        return arguments
    others = []
    placeHolder = "@@"
    for item in [-1, -2]:
        reducedDict = defaultdict(set)
        for path in arguments:
            path = path.strip()
            if path.startswith("/LHCb") or path.startswith("/MC"):
                parsed = path.split("/")
                val = parsed[item]
                parsed[item] = placeHolder
                holderPath = "/".join(parsed)
                reducedDict[holderPath].add(val)
            else:
                others.append(path)
        # Don't allow list of event or file types too long
        arguments = []
        for holderPath, values in reducedDict.items():
            for chunk in breakListIntoChunks(sorted(values), maxList):
                arguments.append(holderPath.replace(placeHolder, ",".join(chunk)))
    # Now sort out the conditions
    conditions = {}
    finalArgs = [arg for arg in arguments if not arg.startswith("/LHCb")]
    for path in [arg for arg in arguments if arg.startswith("/LHCb")]:
        parsed = path.split("/")
        cond = parsed[3]
        parsed[3] = ""
        conditions.setdefault("/".join(parsed), []).append((path, cond))
    for newPath, condTuple in conditions.items():
        if len(condTuple) != 2:
            finalArgs += [path for (path, _c) in condTuple]
        else:
            if len({"-".join(cond.split("-")[0:2])}) == 1:
                finalArgs.append(newPath)
            else:
                finalArgs += [path for (path, _c) in condTuple]

    return sorted(others + finalArgs)


@Script()
def main():
    from DIRAC import gLogger, exit

    Script.registerSwitch("", "NoMerge", "If set, do not merge arguments if BK paths")
    Script.registerSwitch("", "Items=", "Alternative way of passing list of arguments")
    Script.registerSwitch("", "NoParse", "Consider the full lines are arguments and not only the first word")
    Script.registerSwitch("", "Path", "Look for a path in the lines")
    Script.registerSwitch("", "Terse", "Do not print the command executed")
    Script.registerSwitch("", "Last", "Execute with same items as previous call")
    Script.registerSwitch("", "ShowLast", "Show last items used")
    Script.setUsageMessage("\n".join([__doc__]))
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    noMerge = False
    noParse = False
    terse = False
    arguments = []
    path = False
    last = False
    showLast = False
    lastFile = os.path.join(tempfile.gettempdir(), "%d.lastLoops" % os.getppid())
    for switch, val in Script.getUnprocessedSwitches():
        if switch == "NoMerge":
            noMerge = True
        elif switch == "NoParse":
            noParse = True
        elif switch == "Items":
            arguments = val.split(",")
        elif switch == "Terse":
            terse = True
        elif switch == "Path":
            path = True
        elif switch == "Last":
            last = True
        elif switch == "ShowLast":
            showLast = True
            last = True

    if len(args) < 1:
        Script.showHelp(exitCode=1)

    if not arguments:
        if os.path.exists(args[0]):
            oFile = args.pop(0)
        elif last:
            oFile = lastFile
        else:
            oFile = "/dev/stdin"
        with open(oFile) as fd:
            arguments = fd.read().split("\n")
        if showLast:
            gLogger.notice("Last items used:", "\n" + "\n".join(arguments))
            exit(0)

    commands = args

    argList = []
    for arg in arguments:
        if not arg.strip():
            continue
        if path:
            # Look for a path in the line
            # "Real Data" is an exception in BK path (i.e. with space in it)... BKQuery understands without a space
            #    hence for simplicity here, remove the space to allow split() to work ;-)
            words = arg.replace("”", '"').replace("“", '"').replace("Real Data", "RealData").split()
            arg = None
            for word in words:
                if word.startswith("/") or ":/" in word or '"/' in word or "'/" in word:
                    arg = word
                    break
        # If the argument is between quotes, take what is between the quotes
        if not arg:
            continue
        if "'/" in arg:
            arg = arg.split("'")[1]
        elif '"/' in arg:
            arg = arg.split('"')[1]
        elif noParse:
            # Consider the whole line is the argument
            pass
        else:
            # Take the first word
            arg = arg.strip().split()[0]
        # Escape any space left
        if arg:
            argList.append(arg.replace(" ", r"\ "))

    # Persitify items for later use
    if not last and argList:
        with open(lastFile, "w") as tmpFile:
            tmpFile.write("\n".join(argList))

    for arg in reduceArgs(noMerge, argList):
        if arg:
            for command in commands:
                if "@arg@" in command:
                    cmd = command.replace("@arg@", arg, 1) if "dirac-loop" in command else command.replace("@arg@", arg)
                elif command[-1] in ('"', "'"):
                    cmd = command + arg.replace(r"\ ", " ") + command[-1]
                else:
                    cmd = command + " " + arg
                if not terse:
                    gLogger.notice(f"======= {cmd} =========")
                try:
                    output = subprocess.check_output(cmd.encode(), stderr=subprocess.STDOUT, shell=True).decode()
                    gLogger.notice(output[:-1] if terse else output)
                except subprocess.CalledProcessError as e:
                    gLogger.error("Error calling command, return code %d\n" % e.returncode, e.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get Data Quality Flag for the given run."""
import DIRAC

from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            ["Usage:", f"  {Script.scriptName} [option|cfgfile] ... Run ...", "Arguments:", "  Run:      Run number"]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    runSet = {int(id) for arg in Script.getPositionalArgs() for id in arg.split(",")}

    if not runSet:
        Script.showHelp(exitCode=1)

    gLogger.showHeaders(False)
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    cl = BookkeepingClient()

    gLogger.notice("Run Number".ljust(15) + "ExtendedDQOK".ljust(15))

    error = False
    for runId in sorted(runSet):
        retVal = cl.getRunExtendedDQOK(runId)
        if retVal["OK"]:
            systems = ",".join(retVal["Value"])
            gLogger.notice(str(runId).ljust(15) + str(systems).ljust(15))
        else:
            gLogger.error(retVal["Message"])
            error = True

    if error:
        DIRAC.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Gets all Assigned files in a transformation and reports by target SE."""

from collections import defaultdict
from DIRAC.Core.Base.Script import Script


def __getTask(transClient, transID, taskID):
    """Get a task in TS from the transformation and task IDs"""
    res = transClient.getTransformationTasks({"TransformationID": transID, "TaskID": taskID})
    if not res["OK"] or not res["Value"]:
        return None
    return res["Value"][0]


def initStat():
    """Just to initialise a defaultdict to [0,0]"""
    return [0, 0]


@Script()
def main():
    from DIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

    Script.parseCommandLine(ignoreErrors=True)
    transList = getTransformations(Script.getPositionalArgs())

    from DIRAC import gLogger
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    transClient = TransformationClient()

    for transID in transList:
        res = transClient.getTransformationFiles({"TransformationID": transID, "Status": "Assigned"})
        if not res["OK"]:
            gLogger.fatal("Error getting transformation files for %d" % transID)
            continue
        targetStats = defaultdict(initStat)
        taskDict = defaultdict(int)
        for fileDict in res["Value"]:
            taskID = fileDict["TaskID"]
            taskDict[taskID] += 1
        for taskID in taskDict:
            task = __getTask(transClient, transID, taskID)
            targetSE = task.get("TargetSE", None)
            targetStats[targetSE][0] += taskDict[taskID]
            targetStats[targetSE][1] += 1

        gLogger.always("Transformation %d: %d assigned files found" % (transID, len(res["Value"])))
        for targetSE, (nfiles, ntasks) in targetStats.items():
            gLogger.always("\t%s: %d files in %d tasks" % (targetSE, nfiles, ntasks))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import pytest
import json
import os

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise

# WARNING! This code modify specified Oracle instance
# NEVER RUN IT with Production/Certification instance
# WARNING!


def loadJSON(filename: str):
    """Load content from JSON file
    :param str filename: JSON file name
    :returns: python object with content
    """
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        content = json.load(f)
    return content


def clearDBContent(bkdb, content):
    """Cleanup tables specified in the content"""
    for tablename in reversed([table["Table"] for table in content["Tables"]]):
        returnValueOrRaise(bkdb.dbW_.query(f"DELETE FROM {tablename}"))


def fillDBContent(bkdb, content):
    """Load data specified in the content"""
    for table in content["Tables"]:
        tablename = table["Table"]
        for record in table["Records"]:
            columns = list(record)
            sql_fields = ", ".join(columns)
            sql_values = ", ".join([f":{i+1}" for i in range(len(columns))])
            record_values = [record[field] for field in columns]
            sql = f"INSERT INTO {tablename} ({sql_fields}) VALUES ({sql_values})"
            returnValueOrRaise(bkdb.dbW_.query(sql, params=[record_values]))


##############################################


@pytest.fixture
def fileTree(bkdb):
    content = loadJSON("descendents.json")
    clearDBContent(bkdb, content)
    fillDBContent(bkdb, content)
    yield content
    clearDBContent(bkdb, content)


def _orderGetFileDescendents(results):
    success = results["Successful"]
    for lfn in success:
        success[lfn].sort()


def test_file_descendents(bkdb, fileTree):
    result = returnValueOrRaise(
        bkdb.getFileDescendents(["/test/100/1.RAW", "/unknown/file"], depth=5, production=0, checkreplica=False)
    )
    # print(json.dumps(result, sort_keys=True, indent=2)) # to find what it should be, and copy into json...
    expected = fileTree["Results"]["two_5_0_False"]
    assert _orderGetFileDescendents(result) == _orderGetFileDescendents(expected)

    result = returnValueOrRaise(
        bkdb.getFileDescendents(
            ["/test/100/1.RAW", "/test/100/2.RAW", "/test/100/3.RAW"], depth=5, production=0, checkreplica=True
        )
    )
    # print(json.dumps(result, sort_keys=True, indent=2)) # to find what it should be, and copy into json...
    expected = fileTree["Results"]["three_5_0_True"]
    assert _orderGetFileDescendents(result) == _orderGetFileDescendents(expected)

    result = returnValueOrRaise(
        bkdb.getFileDescendents(
            ["/test/100/1.RAW", "/test/100/2.RAW", "/test/100/3.RAW"], depth=5, production=102, checkreplica=True
        )
    )
    # print(json.dumps(result, sort_keys=True, indent=2)) # to find what it should be, and copy into json...
    expected = fileTree["Results"]["three_5_102_True"]
    assert _orderGetFileDescendents(result) == _orderGetFileDescendents(expected)

    result = returnValueOrRaise(
        bkdb.getFileDescendentsTree(
            ["/test/100/1.RAW", "/test/100/2.RAW", "/test/100/3.RAW"], depth=5, production=102, checkreplica=True
        )
    )
    # print(json.dumps(result, sort_keys=True, indent=2)) # to find what it should be, and copy into json...
    expected = fileTree["Results"]["tree_three_5_102_True"]
    assert result == expected



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""BKQuery is a class that decodes BK paths, queries the BK at a high level."""
import os
import sys
from fnmatch import fnmatch
from DIRAC import gLogger
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


def getProcessingPasses(bkQuery, depth=None):
    """Get the list of processing passes for a given BK query The processing pass
    in the initial query may contain "..." or a '*', in which case this acts as a
    wildcard character The search for processing passes may be limited to a
    certain depth (default: all)

    :param bkQuery: BK query dictionary
    :type bkQuery: dict
    :param depth: depth to which look for the processing pass
    :type depth: int
    """
    processingPass = bkQuery.getProcessingPass().replace("...", "*")
    if "*" not in processingPass:
        return [processingPass]
    ind = processingPass.index("*")
    basePass = os.path.dirname(processingPass[:ind])
    bkQuery.setProcessingPass(basePass)
    return sorted(
        pp for pp in bkQuery.getBKProcessingPasses(depth=depth) if fnmatch(pp, processingPass) and pp != basePass
    )


def makeBKPath(bkDict):
    """Builds a path from the dictionary.

    :param bkDict: BK query dictionary
    :type bkDict: dict
    """
    fileType = bkDict.get("FileType", ".")
    if isinstance(fileType, list):
        fileType = ",".join(fileType)
    path = os.path.join(
        "/",
        bkDict.get("ConfigName", ""),
        bkDict.get("ConfigVersion", ""),
        bkDict.get("ConditionDescription", bkDict.get("DataTakingConditions", bkDict.get("SimulationConditions", "."))),
        bkDict.get("ProcessingPass", ".")[1:],
        str(bkDict.get("EventType", ".")).replace("90000000", "."),
        fileType,
    ).replace("/.", "/")
    while path.endswith("/"):
        path = path[:-1]
    return path.replace("RealData", "Real Data")


class BadRunRange(Exception):
    """Exception class for bad run range."""

    pass


def parseRuns(bkQuery, runs):
    """Parse the parameter "Runs" and set it in the dictionary.

    :param bkQuery: BK dictionary
    :type bkQuery: dict
    :param runs: a string, an int, or a list,dict,tuple of run numbers
    :type runs: string or int/long, iterable
    """
    if isinstance(runs, str):
        runs = runs.split(",")
    elif isinstance(runs, (dict, tuple)):
        runs = list(runs)
    elif isinstance(runs, int):
        runs = [str(runs)]
    if len(runs) > 1:
        runList = []
        for run in runs:
            if run.isdigit():
                runList.append(int(run))
            else:
                runRange = run.split(":")
                if len(runRange) == 2 and runRange[0].isdigit() and runRange[1].isdigit():
                    runList += range(int(runRange[0]), int(runRange[1]) + 1)
                else:
                    gLogger.error("Run numbers must be numbers...")
                    raise BadRunRange
        bkQuery["RunNumber"] = runList
    else:
        runs = runs[0].split(":")
        if len(runs) == 1:
            runs = runs[0].split("-")
            if len(runs) == 1:
                bkQuery["RunNumber"] = int(runs[0])
    if "RunNumber" not in bkQuery:
        try:
            if runs[0] and runs[1] and int(runs[0]) > int(runs[1]):
                gLogger.error(f"Warning: End run should be larger than start run: {runs[0]}, {runs[1]}")
                raise BadRunRange
            if runs[0].isdigit():
                bkQuery["StartRun"] = int(runs[0])
            if runs[1].isdigit():
                bkQuery["EndRun"] = int(runs[1])
        except IndexError as ex:  # The runs must be a list
            gLogger.exception("Invalid run range", runs, lException=ex)
            raise BadRunRange
    else:
        if "StartRun" in bkQuery:
            bkQuery.pop("StartRun")
        if "EndRun" in bkQuery:
            bkQuery.pop("EndRun")
    return bkQuery


class BKQuery:
    """It used to build a dictionary using a given Bookkeeping path which is used
    to query the Bookkeeping database."""

    def __init__(self, bkQuery=None, prods=None, runs=None, fileTypes=None, visible=True, eventTypes=None):
        prods = prods if prods is not None else []
        runs = runs if runs is not None else []
        fileTypes = fileTypes if fileTypes is not None else []

        self.extraBKitems = ("StartRun", "EndRun", "Production", "RunNumber")
        self.__bkClient = BookkeepingClient()
        bkPath = ""
        bkQueryDict = {}
        self.__bkFileTypes = set()
        self.__exceptFileTypes = set()
        self.__fakeAllDST = "ZZZZZZZZALL.DST"
        self.__alreadyWarned = False
        if isinstance(bkQuery, BKQuery):
            bkQueryDict = bkQuery.getQueryDict().copy()
        elif isinstance(bkQuery, dict):
            bkQueryDict = bkQuery.copy()
        elif isinstance(bkQuery, str):
            bkPath = bkQuery
        bkQueryDict = self.buildBKQuery(
            bkPath=bkPath,
            bkQueryDict=bkQueryDict,
            prods=prods,
            runs=runs,
            fileTypes=fileTypes,
            eventTypes=eventTypes,
            visible=visible,
        )
        self.__bkPath = bkPath
        self.__bkQueryDict = bkQueryDict
        if not bkQueryDict.get("Visible"):
            self.setVisible(visible)

    def __str__(self):
        return str(self.__bkQueryDict)

    def buildBKQuery(
        self, bkPath="", bkQueryDict=None, prods=None, runs=None, fileTypes=None, visible=True, eventTypes=None
    ):
        """it builds a dictionary using a path."""
        bkQueryDict = bkQueryDict if bkQueryDict is not None else {}
        prods = prods if prods is not None else []
        if not isinstance(prods, list):
            prods = [prods]
        runs = runs if runs is not None else []
        fileTypes = fileTypes if fileTypes is not None else []

        gLogger.verbose(
            f"BKQUERY.buildBKQuery: Path {bkPath}, Dict {bkQueryDict}, Prods {prods}, "
            f"Runs {runs}, FileTypes {fileTypes}, EventTypes {eventTypes}, Visible {visible}"
        )
        self.__bkQueryDict = {}
        if not bkPath and not prods and not bkQueryDict and not runs:
            return {}
        if bkQueryDict:
            bkQuery = bkQueryDict.copy()
        else:
            bkQuery = {}

        # Query given as a path /ConfigName/ConfigVersion/ConditionDescription/ProcessingPass/EventType/FileType
        # or if prefixed with evt: /ConfigName/ConfigVersion/EventType/ConditionDescription/ProcessingPass/FileType
        if bkPath:
            self.__getAllBKFileTypes()
            bkFields = (
                "ConfigName",
                "ConfigVersion",
                "ConditionDescription",
                "ProcessingPass",
                "EventType",
                "FileType",
            )
            url = bkPath.split(":", 1)
            if len(url) == 1:
                bkPath = url[0]
            else:
                if url[0] in ("evt", "evt+std"):
                    bkFields = (
                        "ConfigName",
                        "ConfigVersion",
                        "EventType",
                        "ConditionDescription",
                        "ProcessingPass",
                        "FileType",
                    )
                elif url[0] == "pp":
                    bkFields = ("ProcessingPass", "EventType", "FileType")
                elif url[0] == "prod":
                    bkFields = ("Production", "ProcessingPass", "EventType", "FileType")
                elif url[0] == "runs":
                    bkFields = ("Runs", "ProcessingPass", "EventType", "FileType")
                elif url[0] not in ("sim", "daq", "cond"):
                    gLogger.error(f"Invalid BK path:{bkPath}")
                    return self.__bkQueryDict
                bkPath = url[1]
            if bkPath[0] != "/":
                bkPath = "/" + bkPath
            if bkPath[0:2] == "//":
                bkPath = bkPath[1:]
            bkPath = bkPath.replace("RealData", "Real Data")
            i = 0
            processingPass = "/"
            defaultPP = False
            bk = bkPath.split("/")[1:] + len(bkFields) * [""]
            for bpath in bk:
                gLogger.verbose(
                    "buildBKQuery.1. Item #%d, Field %s, From Path %s, ProcessingPass %s"
                    % (i, bkFields[i], bpath, processingPass)
                )
                if bkFields[i] == "ProcessingPass":
                    # TODO: This should be refactored to be more robust
                    if (
                        bpath != ""
                        and bpath.upper() != "ALL"
                        and not (
                            # Event types are always 8 digits
                            bpath.split(",")[0].split(" ")[0].isdigit()
                            and len(bpath.split(",")[0].split(" ")[0]) >= 8
                        )
                        and not bpath.upper() in self.__bkFileTypes
                    ):
                        processingPass = os.path.join(processingPass, bpath)
                        continue
                    # Set the PP
                    if processingPass != "/":
                        bkQuery["ProcessingPass"] = processingPass
                    else:
                        defaultPP = True
                    i += 1
                gLogger.verbose(
                    "buildBKQuery.2. Item #%d, Field %s, From Path %s, ProcessingPass %s"
                    % (i, bkFields[i], bpath, processingPass)
                )
                if bkFields[i] == "EventType" and bpath:
                    eventTypeList = []
                    # print b
                    if bpath.upper() == "ALL":
                        bpath = "ALL"
                    else:
                        for et in bpath.split(","):
                            try:
                                eventType = int(et.split(" ")[0])
                                eventTypeList.append(eventType)
                            except ValueError:
                                pass
                        if len(eventTypeList) == 1:
                            eventTypeList = eventTypeList[0]
                        bpath = eventTypeList
                        gLogger.verbose(f"buildBKQuery. Event types {eventTypeList}")
                # Set the BK dictionary item
                if bpath != "":
                    bkQuery[bkFields[i]] = bpath
                if defaultPP:
                    # PP was empty, try once more to get the Event Type
                    defaultPP = False
                else:
                    # Go to next item
                    i += 1
                if i == len(bkFields):
                    break

            gLogger.verbose(f"buildBKQuery. Query dict {str(bkQuery)}")
            # Set default event type to real data
            if bkQuery.get("ConfigName") != "MC" and not bkQuery.get("EventType"):
                bkQuery["EventType"] = "90000000"
            if bkQuery.get("EventType") == "ALL":
                bkQuery.pop("EventType")

        # Run limits are given
        runs = bkQuery.pop("Runs", runs)
        if runs:
            try:
                bkQuery = parseRuns(bkQuery, runs)
            except BadRunRange:
                return self.__bkQueryDict

        # Query given as a list of production
        if prods and str(prods[0]).upper() != "ALL":
            try:
                bkQuery.setdefault("Production", []).extend([int(prod) for prod in prods])
            except ValueError as ex:  # The prods list does not contains numbers
                gLogger.warn(ex)
                gLogger.error("Invalid production list", str(prods))
                return self.__bkQueryDict

        # If an event type is specified
        if eventTypes:
            bkQuery["EventType"] = eventTypes

        # Set the file type(s) taking into account excludes file types
        fileTypes = bkQuery.get("FileType", fileTypes)
        bkQuery.pop("FileType", None)
        self.__bkQueryDict = bkQuery.copy()
        fileType = self.__fileType(fileTypes)
        # print fileType
        if fileType:
            bkQuery["FileType"] = fileType

        # Remove all "ALL"'s in the dict, if any
        for i in self.__bkQueryDict:
            if isinstance(bkQuery[i], str) and bkQuery[i] == "ALL":
                bkQuery.pop(i)

        # If there is only one production, make it faster with a single value rather than a list
        prodList = bkQuery.get("Production")
        if isinstance(prodList, list) and len(prodList) == 1:
            bkQuery["Production"] = prodList[0]
        self.__bkQueryDict = bkQuery.copy()
        self.setVisible(visible)

        # Set both event type entries
        # print "Before setEventType", self.__bkQueryDict
        if not self.setEventType(bkQuery.get("EventType")):
            self.__bkQueryDict = {}
            return self.__bkQueryDict
        # Set conditions
        # print "Before setConditions", self.__bkQueryDict
        self.setConditions(
            bkQuery.get(
                "ConditionDescription", bkQuery.get("DataTakingConditions", bkQuery.get("SimulationConditions"))
            )
        )
        # print "Returned value", self.__bkQueryDict
        return self.__bkQueryDict

    def setOption(self, key, val):
        """It insert an item to the dictionary.

        The key is an bookkeeping attribute (condition).
        """
        if val:
            self.__bkQueryDict[key] = val
        else:
            self.__bkQueryDict.pop(key, None)
        return self.__bkQueryDict

    def setConditions(self, cond=None):
        """Set the dictionary items for a given condition, or remove it
        (cond=None)"""
        if "ConfigName" not in self.__bkQueryDict and cond:
            gLogger.warn("Impossible to set Conditions to a BK Query without Configuration")
            return self.__bkQueryDict
        # There are two items in the dictionary: ConditionDescription and Simulation/DataTaking-Conditions
        eventType = self.__bkQueryDict.get("EventType", "ALL")
        if self.__bkQueryDict.get("ConfigName") == "MC" or (
            isinstance(eventType, str) and eventType.upper() != "ALL" and eventType[0] != "9"
        ):
            conditionsKey = "SimulationConditions"
        else:
            conditionsKey = "DataTakingConditions"
        self.setOption("ConditionDescription", cond)
        return self.setOption(conditionsKey, cond)

    def setFileType(self, fileTypes=None):
        """insert the file type to the Boookkeeping dictionary."""
        return self.setOption("FileType", self.__fileType(fileTypes))

    def setDQFlag(self, dqFlag="OK"):
        """Sets the data quality."""
        if isinstance(dqFlag, str):
            dqFlag = dqFlag.upper()
        elif isinstance(dqFlag, list):
            dqFlag = [dq.upper() for dq in dqFlag]
        return self.setOption("DataQuality", dqFlag)

    def setStartDate(self, startDate):
        """Sets the start date."""
        return self.setOption("StartDate", startDate)

    def setEndDate(self, endDate):
        """Sets the end date."""
        return self.setOption("EndDate", endDate)

    def setProcessingPass(self, processingPass):
        """Sets the processing pass."""
        return self.setOption("ProcessingPass", processingPass)

    def setEventType(self, eventTypes=None):
        """Sets the event type."""
        if eventTypes:
            if isinstance(eventTypes, str):
                eventTypes = eventTypes.split(",")
            elif not isinstance(eventTypes, list):
                eventTypes = [eventTypes]
            try:
                eventTypes = [str(int(et)) for et in eventTypes]
            except ValueError as ex:
                gLogger.warn(ex)
                gLogger.error("Invalid list of event types", eventTypes)
                return {}
            if isinstance(eventTypes, list) and len(eventTypes) == 1:
                eventTypes = eventTypes[0]
        return self.setOption("EventType", eventTypes)

    def setVisible(self, visible=None):
        """Sets the visibility flag."""
        if visible is True or (isinstance(visible, str) and visible[0].lower() == "y"):
            visible = "Yes"
        if visible is False:
            visible = "No"
        return self.setOption("Visible", visible)

    def setExceptFileTypes(self, fileTypes):
        """Sets the expected file types."""
        if not isinstance(fileTypes, list):
            fileTypes = [fileTypes]
        self.__exceptFileTypes.update(fileTypes)
        self.setFileType([t for t in self.getFileTypeList() if t not in fileTypes])

    def getExceptFileTypes(self):
        return list(self.__exceptFileTypes)

    def getQueryDict(self):
        """Returns the bookkeeping dictionary."""
        return self.__bkQueryDict

    def getPath(self):
        """Returns the Bookkeeping path."""
        return self.__bkPath

    def makePath(self):
        """Builds a path from the dictionary."""
        bk = self.__bkQueryDict
        fileType = bk.get("FileType", "")
        if isinstance(fileType, list):
            fileType = ",".join(fileType)
        path = os.path.join(
            "/",
            bk.get("ConfigName", ""),
            bk.get("ConfigVersion", ""),
            bk.get("ConditionDescription", "."),
            bk.get("ProcessingPass", ".")[1:],
            str(bk.get("EventType", ".")).replace("90000000", "."),
            fileType,
        ).replace("/./", "//")
        while True:
            if path.endswith("/"):
                path = path[:-1]
            else:
                return path

    def getFileTypeList(self):
        """Returns the file types."""
        fileTypes = self.__bkQueryDict.get("FileType", [])
        if not isinstance(fileTypes, list):
            fileTypes = [fileTypes]
        return fileTypes

    def getEventTypeList(self):
        """Returns the event types."""
        eventType = self.__bkQueryDict.get("EventType", [])
        if eventType:
            if not isinstance(eventType, list):
                eventType = [eventType]
        return eventType

    def getProcessingPass(self):
        """Returns the processing pass."""
        return self.__bkQueryDict.get("ProcessingPass", "")

    def getConditions(self):
        """Returns the Simulation/data taking conditions."""
        return self.__bkQueryDict.get("ConditionDescription", "")

    def getConfiguration(self):
        """Returns the configuration name and configuration version."""
        configName = self.__bkQueryDict.get("ConfigName", "")
        configVersion = self.__bkQueryDict.get("ConfigVersion", "")
        if not configName or not configVersion:
            return ""
        return os.path.join("/", configName, configVersion)

    def isVisible(self):
        """Returns True/False depending on the visibility flag."""
        return self.__bkQueryDict.get("Visible", "All")

    def __fileType(self, fileType=None, returnList=False):
        """return the file types taking into account the expected file types."""
        gLogger.verbose(f"BKQuery.__fileType: {self}, fileType: {fileType}")
        if not fileType:
            return []
        self.__getAllBKFileTypes()
        if isinstance(fileType, list):
            fileTypes = fileType
        else:
            fileTypes = fileType.split(",")
        allRequested = None
        if fileTypes[0].lower() == "all":
            allRequested = True
            bkTypes = self.getBKFileTypes()
            gLogger.verbose(f"BKQuery.__fileType: bkTypes {str(bkTypes)}")
            if bkTypes:
                fileTypes = list(set(bkTypes) - self.__exceptFileTypes)
            else:
                fileTypes = []
        expandedTypes = set()
        # print "Requested", fileTypes
        for ft in fileTypes:
            if ft.lower() == "all.hist":
                allRequested = False
                expandedTypes.update(
                    [t for t in self.__exceptFileTypes.union(self.__bkFileTypes) if t.endswith("HIST")]
                )
            elif ft.lower().find("all.") == 0:
                ext = "." + ft.split(".")[1]
                ft = []
                if allRequested is None:
                    allRequested = True
                expandedTypes.update(
                    [t for t in set(self.getBKFileTypes()) - self.__exceptFileTypes if t.endswith(ext)]
                )
            else:
                expandedTypes.add(ft)
        # Remove __exceptFileTypes only if not explicitly required
        # print "Obtained", fileTypes, expandedTypes
        gLogger.verbose(
            f"BKQuery.__fileType: requested {allRequested}, expanded {expandedTypes}, except {self.__alreadyWarned}"
        )
        if expandedTypes - self.__bkFileTypes and not self.__alreadyWarned:
            self.__alreadyWarned = True
            gLogger.always(
                "**** Take care: some requested file types do not exist!!",
                str(sorted(expandedTypes - self.__bkFileTypes)),
            )
        if allRequested or not expandedTypes & self.__exceptFileTypes:
            expandedTypes -= self.__exceptFileTypes
        gLogger.verbose(f"BKQuery.__fileType: result {sorted(expandedTypes)}")
        if len(expandedTypes) == 1 and not returnList:
            return list(expandedTypes)[0]
        return list(expandedTypes)

    def __getAllBKFileTypes(self):
        """Returns the file types from the bookkeeping database."""
        if not self.__bkFileTypes:
            self.__bkFileTypes = {self.__fakeAllDST}
            warned = False
            while True:
                res = self.__bkClient.getAvailableFileTypes()
                if res["OK"]:
                    dbresult = res["Value"]
                    for record in dbresult["Records"]:
                        if (
                            record[0].endswith("HIST")
                            or record[0].endswith("ETC")
                            or record[0] == "LOG"
                            or record[0].endswith("ROOT")
                        ):
                            self.__exceptFileTypes.add(record[0])
                        self.__bkFileTypes.add(record[0])
                    break
                if not warned:
                    gLogger.always("Error getting BK file types, retrying", res["Message"])
                    warned = True

    def __getBKFiles(self, bkQueryDict, retries=5):
        """Call BK getFiles() with some retries."""
        if not retries:
            retries = sys.maxsize
        errorLogged = False
        while retries:
            res = self.__bkClient.getFiles(bkQueryDict)
            if res["OK"]:
                break
            retries -= 1
            if not errorLogged:
                errorLogged = True
                gLogger.warn("Error getting files from BK, retrying...", res["Message"])
        return res

    def getLFNsAndSize(self, getSize=True):
        """Returns the LFNs and their size for a given data set."""
        self.__getAllBKFileTypes()
        res = self.__getBKFiles(self.__bkQueryDict)
        lfns = []
        lfnSize = 0
        if not res["OK"]:
            gLogger.error(f"Error from BK for {self.__bkQueryDict}:", res["Message"])
        else:
            lfns = set(res["Value"])
            exceptFiles = list(self.__exceptFileTypes)
            if exceptFiles and not self.__bkQueryDict.get("FileType"):
                res = self.__getBKFiles(BKQuery(self.__bkQueryDict).setOption("FileType", exceptFiles))
                if res["OK"]:
                    lfnsExcept = set(res["Value"]) & lfns
                else:
                    gLogger.error(
                        f"***** ERROR ***** Error in getting dataset from BK for {exceptFiles} files:", res["Message"]
                    )
                    lfnsExcept = set()
                if lfnsExcept:
                    gLogger.warn(
                        f"***** WARNING ***** Found {len(lfnsExcept)} files in BK query that will be excluded!"
                    )
                    gLogger.verbose("Exclusion file types", str(exceptFiles))
                    gLogger.warn("                    If creating a transformation, set '--FileType ALL'")
                    lfns = lfns - lfnsExcept
                else:
                    exceptFiles = False
            if getSize:
                # Get size only if needed
                query = BKQuery(self.__bkQueryDict)
                query.setOption("FileSize", True)
                res = self.__getBKFiles(query.getQueryDict())
                if res["OK"] and isinstance(res["Value"], list) and res["Value"][0]:
                    lfnSize = res["Value"][0]
                if exceptFiles and not self.__bkQueryDict.get("FileType"):
                    res = self.__getBKFiles(query.setOption("FileType", exceptFiles))
                    if res["OK"] and isinstance(res["Value"], list) and res["Value"][0]:
                        lfnSize -= res["Value"][0]

                lfnSize /= 1e12  # 1 Tera
            else:
                lfnSize = 0.0
        return {"LFNs": list(lfns), "LFNSize": lfnSize}

    def getLFNSize(self, visible=None):
        """Returns the size of a  given data set."""
        if visible is None:
            visible = self.isVisible()
        res = self.__getBKFiles(BKQuery(self.__bkQueryDict, visible=visible).setOption("FileSize", True))
        if res["OK"] and isinstance(res["Value"], list) and res["Value"][0]:
            lfnSize = res["Value"][0]
        else:
            lfnSize = 0
        return lfnSize

    def getNumberOfLFNs(self, visible=None):
        """Returns the number of LFNs correspond to a given data set."""
        if visible is None:
            visible = self.isVisible()
        if self.isVisible() != visible:
            query = BKQuery(self.__bkQueryDict, visible=visible)
        else:
            query = self
        fileTypes = query.getFileTypeList()
        nbFiles = 0
        size = 0

        for ft in fileTypes:
            if ft:
                res = self.__bkClient.getFilesSummary(query.setFileType(ft))
                # print query, res
                if res["OK"]:
                    res = res["Value"]
                    ind = res["ParameterNames"].index("NbofFiles")
                    if res["Records"][0][ind]:
                        nbFiles += res["Records"][0][ind]
                        ind1 = res["ParameterNames"].index("FileSize")
                        size += res["Records"][0][ind1]
                        # print 'Visible',query.isVisible(),ft, 'Files:',
                        # res['Records'][0][ind], 'Size:', res['Records'][0][ind1]
                else:
                    gLogger.error("Error getting files summary", res["Message"])
                    break
        return {"NumberOfLFNs": nbFiles, "LFNSize": size}

    def getLFNs(self, printSEUsage=False, printOutput=True, visible=None):
        """returns a list of lfns.

        It prints statistics about the data sets if it is requested.
        """
        if visible is None:
            visible = self.isVisible()

        if self.isVisible() != visible:
            query = BKQuery(self.__bkQueryDict, visible=visible)
        else:
            query = self

        # Loop for each production or each event type rather than make a single query
        loopItem = None
        loopList = []
        prods = self.__bkQueryDict.get("Production")
        eventTypes = self.__bkQueryDict.get("EventType")
        if prods and isinstance(prods, list):
            loopItem = "Production"
            loopList = prods
        elif eventTypes and isinstance(eventTypes, list):
            loopItem = "EventType"
            loopList = eventTypes

        if loopItem:
            # It's faster to loop on a list of prods or event types than query the BK with a list as argument
            lfns = []
            lfnSize = 0
            if query == self:
                query = BKQuery(self.__bkQueryDict, visible=visible)
            for item in loopList:
                query.setOption(loopItem, item)
                lfnsAndSize = query.getLFNsAndSize(getSize=printOutput)
                lfns += lfnsAndSize["LFNs"]
                lfnSize += lfnsAndSize["LFNSize"]
        else:
            lfnsAndSize = query.getLFNsAndSize(getSize=printOutput)
            lfns = lfnsAndSize["LFNs"]
            lfnSize = lfnsAndSize["LFNSize"]

        if not lfns:
            gLogger.verbose("No files found for BK query", self.__bkQueryDict)
        else:
            lfns.sort()

            # Only for printing
            if printOutput:
                gLogger.notice(f"\n{len(lfns)} files ({lfnSize:.1f} TB) in directories:")
                dirs = {}
                for lfn in lfns:
                    directory = os.path.join(os.path.dirname(lfn), "")
                    dirs[directory] = dirs.setdefault(directory, 0) + 1
                for directory in sorted(dirs):
                    gLogger.notice(f"{directory} {dirs[directory]} files")
                if printSEUsage:
                    rpc = StorageUsageClient()
                    totalUsage = {}
                    totalSize = 0
                    for directory in dirs:
                        res = rpc.getStorageSummary(directory, "", "", [])
                        if res["OK"]:
                            for se in [se for se in res["Value"] if not se.endswith("-ARCHIVE")]:
                                totalUsage[se] = totalUsage.setdefault(se, 0) + res["Value"][se]["Size"]
                                totalSize += res["Value"][se]["Size"]
                    ses = sorted(totalUsage)
                    totalUsage["Total"] = totalSize
                    ses.append("Total")
                    gLogger.notice(f"\n{'SE'.ljust(20)} Size (TB)")
                    for se in ses:
                        gLogger.notice(f"{se.ljust(20)} {totalUsage[se] / 1000000000000.0:.1f}")
        return lfns

    def getDirs(self, printOutput=False, visible=None):
        """Returns the directories."""
        if visible is None:
            visible = self.isVisible()
        lfns = self.getLFNs(printSEUsage=True, printOutput=printOutput, visible=visible)
        dirs = set()
        for lfn in lfns:
            dirs.add(os.path.dirname(lfn))
        return sorted(dirs)

    @staticmethod
    def __getProdStatus(prod):
        """Returns the status of a given transformation."""
        res = TransformationClient().getTransformation(prod, extraParams=False)
        if not res["OK"]:
            gLogger.error("Could not get information", f"on production {prod}")
            return None
        return res["Value"]["Status"]

    def getBKRuns(self):
        """It returns a list of runs from the bookkeeping."""
        if self.getProcessingPass().replace("/", "") == "Real Data":
            return self.getBKProductions()

    def getBKProductions(self, visible=None):
        """It returns a list of productions."""
        if visible is None:
            visible = self.isVisible()
        prodList = self.__bkQueryDict.get("Production")
        if prodList:
            if not isinstance(prodList, list):
                prodList = [prodList]
            return sorted(prodList)
        #        It seems this protection is not really necessary and was preventing another script to work
        #        if not self.getProcessingPass():
        #            gLogger.fatal("Impossible to get a list of productions without the Processing Pass")
        #            return []
        eventTypes = self.__bkQueryDict.get("EventType")
        if not isinstance(eventTypes, list):
            eventTypes = [eventTypes]
        fullList = set()
        for eventType in eventTypes:
            bkQ = BKQuery(self.__bkQueryDict)
            bkQ.setVisible(visible)
            bkDict = bkQ.setEventType(eventType)
            # gLogger.notice( 'Get productions for BK query', str( bkDict ) )
            res = self.__bkClient.getProductions(bkDict)
            if not res["OK"]:
                gLogger.error("Error getting productions from BK", res["Message"])
                return []
            if self.getProcessingPass().replace("/", "") != "Real Data":
                fileTypes = self.getFileTypeList()
                prodList = {
                    prod
                    for prods in res["Value"]["Records"]
                    for prod in prods
                    if self.__getProdStatus(prod) != "Deleted"
                }
                # print '\n', self.__bkQueryDict, res['Value']['Records'], '\nVisible:', visible, prodList
                pList = set()
                if fileTypes:
                    transClient = TransformationClient()
                    for prod in prodList:
                        res = transClient.getBookkeepingQuery(prod)
                        if res["OK"] and res["Value"]["FileType"] in fileTypes:
                            pList.add(prod)
                if not pList:
                    pList = prodList
            else:
                runList = sorted(-run for r in res["Value"]["Records"] for run in r)
                startRun = int(self.__bkQueryDict.get("StartRun", 0))
                endRun = int(self.__bkQueryDict.get("EndRun", sys.maxsize))
                pList = {run for run in runList if run >= startRun and run <= endRun}
            fullList.update(pList)
        return sorted(fullList)

    def getBKConditions(self):
        """It returns the data taking / simulation conditions."""
        conditions = self.__bkQueryDict.get("ConditionDescription")
        if conditions:
            if not isinstance(conditions, list):
                conditions = [conditions]
            return conditions
        result = self.__bkClient.getConditions(self.__bkQueryDict)
        if result["OK"]:
            resList = result["Value"]
        else:
            return []
        conditions = []
        for res in resList:
            ind = res["ParameterNames"].index("Description")
            if res["Records"]:
                conditions += [par[ind] for par in res["Records"]]
                break
        return sorted(conditions)

    def getBKEventTypes(self):
        """It returns the event types."""
        eventType = self.getEventTypeList()
        if eventType:
            return eventType
        res = self.__bkClient.getEventTypes(self.__bkQueryDict)["Value"]
        ind = res["ParameterNames"].index("EventType")
        eventTypes = sorted(rec[ind] for rec in res["Records"])
        return eventTypes

    def getBKFileTypes(self, bkDict=None):
        """It returns the file types."""
        fileTypes = self.getFileTypeList()
        # print "Call getBKFileTypes:", self, fileTypes
        if not fileTypes:
            if not bkDict:
                bkDict = self.__bkQueryDict.copy()
            else:
                bkDict = bkDict.copy()
            bkDict.setdefault("Visible", "All")
            bkDict["ReplicaFlag"] = "All"
            bkDict.pop("RunNumber", None)
            fileTypes = []
            eventTypes = bkDict.get("EventType")
            if isinstance(eventTypes, list):
                for et in eventTypes:
                    bkDict["EventType"] = et
                    fileTypes += self.getBKFileTypes(bkDict)
            else:
                res = self.__bkClient.getFileTypes(bkDict)
                if res["OK"]:
                    res = res["Value"]
                    ind = res["ParameterNames"].index("FileTypes")
                    fileTypes = [rec[ind] for rec in res["Records"] if rec[ind] not in self.__exceptFileTypes]
        if "ALL.DST" in fileTypes:
            fileTypes.remove("ALL.DST")
            fileTypes.append(self.__fakeAllDST)
        # print 'FileTypes1', fileTypes
        fileTypes = self.__fileType(fileTypes, returnList=True)
        # print 'FileTypes2', fileTypes
        if self.__fakeAllDST in fileTypes:
            fileTypes.remove(self.__fakeAllDST)
            fileTypes.append("ALL.DST")
        # print 'FileTypes3', fileTypes
        return fileTypes

    def getBKProcessingPasses(self, queryDict=None, depth=None):
        """It returns the processing pass."""
        if depth is None:
            depth = sys.maxsize
        processingPasses = {}
        if not queryDict:
            queryDict = self.__bkQueryDict.copy()
        initialPP = queryDict.get("ProcessingPass", "/")
        # print "Start", initialPP, queryDict
        res = self.__bkClient.getProcessingPass(queryDict, initialPP)
        if not res["OK"]:
            if "Empty Directory" not in res["Message"]:
                gLogger.error(f"ERROR getting processing passes for {queryDict}", res["Message"])
            return {}
        ppRecords = res["Value"][0]
        if "Name" in ppRecords["ParameterNames"]:
            ind = ppRecords["ParameterNames"].index("Name")
            passes = sorted(os.path.join(initialPP, rec[ind]) for rec in ppRecords["Records"] if rec[ind] is not None)
        else:
            passes = []
        evtRecords = res["Value"][1]
        if "EventType" in evtRecords["ParameterNames"]:
            ind = evtRecords["ParameterNames"].index("EventType")
            eventTypes = [str(rec[ind]) for rec in evtRecords["Records"]]
        else:
            eventTypes = []

        if passes and depth:
            depth -= 1
            nextProcessingPasses = {}
            for pName in passes:
                processingPasses[pName] = []
                queryDict["ProcessingPass"] = pName
                nextProcessingPasses.update(self.getBKProcessingPasses(queryDict, depth=depth))
            processingPasses.update(nextProcessingPasses)
        if eventTypes:
            processingPasses[initialPP] = eventTypes
        for pName in ("/Real Data", "/"):
            if pName in processingPasses:
                processingPasses.pop(pName)
        # print "End", initialPP, [(key, processingPasses[key]) for key in sorted(list(processingPasses))]
        return processingPasses



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import datetime
import pytest
from xml.dom.minidom import parseString
from unittest.mock import MagicMock

mockBKDB = MagicMock()
mockBKDB.return_value = None

from DIRAC import gLogger

gLogger.setLevel("DEBUG")

# sut
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.JobReader import JobReader
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.XMLFilesReaderManager import XMLFilesReaderManager

xmlString = (
    """<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE Job SYSTEM "book.dtd">
<Job ConfigName="test" ConfigVersion="Jenkins" Date="%jDate%" Time="%jTime%">
  <TypedParameter Name="CPUTIME" Type="Info" Value="111222"/>
  <TypedParameter Name="ExecTime" Type="Info" Value="36571.0480781"/>
  <TypedParameter Name="WNMODEL" Type="Info" Value="Intel(R)Xeon(R)CPUE5-2650v2@2.60GHz"/>
  <TypedParameter Name="WNCPUPOWER" Type="Info" Value="1"/>
  <TypedParameter Name="WNCACHE" Type="Info" Value="2593.748"/>
  <TypedParameter Name="WorkerNode" Type="Info" Value="b6bd1ec9ae.cern.ch"/>
  <TypedParameter Name="WNMEMORY" Type="Info" Value="1667656.0"/>
  <TypedParameter Name="WNCPUHS06" Type="Info" Value="11.4"/>
  <TypedParameter Name="Production" Type="Info" Value="12345"/>
  <TypedParameter Name="DiracJobId" Type="Info" Value="147844677"/>
  <TypedParameter Name="Name" Type="Info" Value="00056438_00001025_test_1"/>
  <TypedParameter Name="JobStart" Type="Info" Value="%jStart%"/>
  <TypedParameter Name="JobEnd" Type="Info" Value="%jEnd%"/>
  <TypedParameter Name="Location" Type="Info" Value="LCG.CERN.ch"/>
  <TypedParameter Name="JobType" Type="Info" Value="MCSimulation"/>
  <TypedParameter Name="ProgramName" Type="Info" Value="Gauss"/>
  <TypedParameter Name="ProgramVersion" Type="Info" Value="v49r5"/>
  <TypedParameter Name="DiracVersion" Type="Info" Value="v6r15p9"/>
  <TypedParameter Name="FirstEventNumber" Type="Info" Value="1"/>
  <TypedParameter Name="StatisticsRequested" Type="Info" Value="-1"/>
  <TypedParameter Name="StepID" Type="Info" Value="123"/>
  <TypedParameter Name="NumberOfEvents" Type="Info" Value="411"/>
  <OutputFile Name="/lhcb/MC/2012/SIM/00056438/0000/00056438_00001025_test_1.sim" TypeName="SIM" TypeVersion="ROOT">
          <Parameter Name="EventTypeId" Value="11104131"/>
          <Parameter Name="EventStat" Value="411"/>
          <Parameter Name="FileSize" Value="862802861"/>
          <Parameter Name="MD5Sum" Value="ae647981ea419cc9f8e8fa0a2d6bfd3d"/>
          <Parameter Name="Guid" Value="546014C4-55C6-E611-8E94-02163E00F6B2"/>
  </OutputFile>
  <OutputFile Name="/lhcb/MC/2012/LOG/00056438/0000/00001025/Gauss_00056438_00001025_test_1.log" """
    + """TypeName="LOG" TypeVersion="1">
          <Parameter Name="FileSize" Value="319867"/>
          <Replica Location="Web" Name="http://lhcb-logs.cern.ch/"""
    + """storage/lhcb/MC/2012/LOG/00056438/0000/00001025/Gauss_00056438_00001025_test_1.log"/>
          <Parameter Name="MD5Sum" Value="e4574c9083d1163d43ba6ac033cbd769"/>
          <Parameter Name="Guid" Value="E4574C90-83D1-163D-43BA-6AC033CBD769"/>
  </OutputFile>
  <SimulationCondition>
          <Parameter Name="SimDescription" Value="Beam4000GeV-2012-MagUp-Nu2.5-Pythia8"/>
  </SimulationCondition>
</Job>
"""
)


def test_JobReader():
    currentTime = datetime.datetime.now()
    jobStart = jobEnd = datetime.datetime.now()
    jobStart = jobEnd = jobStart.replace(second=0, microsecond=0)
    xml = xmlString.replace("%jDate%", currentTime.strftime("%Y-%m-%d"))
    xml = xml.replace("%jTime%", currentTime.strftime("%H:%M"))
    xml = xml.replace("%jStart%", jobStart.strftime("%Y-%m-%d %H:%M"))
    xml = xml.replace("%jEnd%", jobEnd.strftime("%Y-%m-%d %H:%M"))
    doc = parseString(xml)

    job = JobReader().readJob(doc, "IN Memory")
    assert job.configuration.configName == "test"
    assert job.configuration.configVersion == "Jenkins"
    assert len(job.outputFiles) == 2
    assert job.outputFiles[0].name == "/lhcb/MC/2012/SIM/00056438/0000/00056438_00001025_test_1.sim"
    assert job.outputFiles[0].type == "SIM"
    assert job.outputFiles[0].params[1].value == "411"
    assert len(job.parameters) == 22
    assert job.parameters[0].name == "CPUTIME"
    assert job.parameters[0].value == "111222"
    assert job.simulationCondition.parameters["SimDescription"] == "Beam4000GeV-2012-MagUp-Nu2.5-Pythia8"


@pytest.mark.parametrize(
    "inputfiles, getRunNbAndTckRV, expected",
    [
        ([], {"OK": True}, {"OK": True, "Value": (set(), set())}),
        (["aa"], {"OK": True, "Value": [(None, "None")]}, {"OK": True, "Value": (set(), set())}),
        (["aa"], {"OK": False, "Message": "bof"}, {"OK": False, "Message": "bof"}),
        (["aa", "bb"], {"OK": True, "Value": [(123, "None")]}, {"OK": True, "Value": ({123}, set())}),
        (["aa", "bb"], {"OK": True, "Value": [(123, "x123")]}, {"OK": True, "Value": ({123}, {"x123"})}),
    ],
)
def test__getRunNumbersAndTCKs(mocker, inputfiles, getRunNbAndTckRV, expected):
    mocker.patch(
        "LHCbDIRAC.BookkeepingSystem.Service.XMLReader.XMLFilesReaderManager.OracleBookkeepingDB.__init__",
        side_effect=mockBKDB,
    )
    xfrm = XMLFilesReaderManager()
    xfrm.db = MagicMock()
    xfrm.db.getRunNbAndTck.return_value = getRunNbAndTckRV

    res = xfrm._getRunNumbersAndTCKs(inputfiles)
    assert res == expected


@pytest.mark.parametrize(
    "prod, runNumber, getProductionProcessingPassID_RV, getRunAndProcessingPassDataQuality_RV, expected",
    [
        #  no RunNumber -> OK, None
        (None, None, {"OK": True}, {"OK": True}, {"OK": True, "Value": None}),
        # No processing pass -> OK, None
        (None, 123, {"OK": True, "Value": None}, {"OK": True}, {"OK": True, "Value": None}),
        # No DQ -> OK, None
        (None, 123, {"OK": True, "Value": 1}, {"OK": True, "Value": None}, {"OK": True, "Value": None}),
        # All OK -> OK, DQ value
        (None, 123, {"OK": True, "Value": 1}, {"OK": True, "Value": "DQValue"}, {"OK": True, "Value": "DQValue"}),
        # processing pass ERROR -> ERROR
        (None, 123, {"OK": False, "Message": "NOK"}, {"OK": True, "Value": "OK"}, {"OK": False, "Message": "NOK"}),
        # DQ Error -> OK, None (not an error, see method doc)
        (None, 123, {"OK": True, "Value": "OK"}, {"OK": False, "Message": "NOK"}, {"OK": True, "Value": None}),
        # All OK -> OK, DQValue
        (321, 123, {"OK": True, "Value": 1}, {"OK": True, "Value": "UNCHECKED"}, {"OK": True, "Value": "UNCHECKED"}),
    ],
)
def test__getDataQuality(
    mocker, prod, runNumber, getProductionProcessingPassID_RV, getRunAndProcessingPassDataQuality_RV, expected
):
    mocker.patch(
        "LHCbDIRAC.BookkeepingSystem.Service.XMLReader.XMLFilesReaderManager.OracleBookkeepingDB.__init__",
        side_effect=mockBKDB,
    )
    xfrm = XMLFilesReaderManager()
    xfrm.db = MagicMock()
    xfrm.db.getProductionProcessingPassID.return_value = getProductionProcessingPassID_RV
    xfrm.db.getRunAndProcessingPassDataQuality.return_value = getRunAndProcessingPassDataQuality_RV

    res = xfrm._getDataQuality(prod, runNumber)
    assert res == expected


def test_SMOG_Parser():
    assert XMLFilesReaderManager.smogInjectionSMOG2Gas("NONE") == "NoGas"
    assert XMLFilesReaderManager.smogInjectionSMOG2Gas("NONE_NONE") == "NoGas"
    assert XMLFilesReaderManager.smogInjectionSMOG2Gas("SMOG_ARGON") == "Unknown"
    assert XMLFilesReaderManager.smogInjectionSMOG2Gas("SMOG2_ARGON") == "Argon"
    assert XMLFilesReaderManager.smogInjectionSMOG2Gas("SMOG_HELIUM_Unstable") == "Unknown"
    assert XMLFilesReaderManager.smogInjectionSMOG2Gas("SMOG2_HELIUM_Unstable") == "HeliumUnstable"



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Utility for invoking running LHCb applications."""
import asyncio
import json
import shlex
import subprocess

from DIRAC import gLogger, gConfig, S_OK
from DIRAC.WorkloadManagementSystem.Utilities.RemoteRunner import RemoteRunner
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import getEventsToProduce


class LHCbApplicationError(RuntimeError):
    """Exception for application errors."""


class LHCbDIRACError(RuntimeError):
    """Exception for application errors."""


class RunApplication:
    """Encapsulate logic for running an LHCb application."""

    def __init__(
        self,
        gaudiAppModule,
        commandOptions,
        stepOutputTypes,
        histogram,
        runNumberGauss,
        firstEventNumberGauss,
        eventTimeout,
    ):
        self.log = gLogger.getSubLogger("RunApplication")

        self.applicationName = gaudiAppModule.applicationName
        self.cleanedApplicationName = gaudiAppModule.cleanedApplicationName
        self.applicationVersion = gaudiAppModule.applicationVersion
        self.prmonPath = "/cvmfs/lhcb.cern.ch/lib/experimental/prmon/bin/prmon"
        self.usePrmon = gaudiAppModule.usePrmon

        if gaudiAppModule.executable != "gaudirun.py":
            raise LHCbDIRACError(f"gaudiAppModule.executable={gaudiAppModule.executable!r} is not supported")

        self.prodConfFileName = "prodConf_{}_{}_{}_{}.json".format(
            self.cleanedApplicationName,
            gaudiAppModule.production_id,
            gaudiAppModule.prod_job_id,
            gaudiAppModule.step_number,
        )
        self.numberOfProcessors = gaudiAppModule.numberOfProcessors
        self.applicationLog = gaudiAppModule.applicationLog or "applicationLog.txt"
        self.stdError = gaudiAppModule.stdError or self.applicationLog

        # Sanity checks
        if not gaudiAppModule.stepInputData and self.applicationName.lower() != "gauss":
            raise RuntimeError("No MC, but no input data")
        if gaudiAppModule.TCK and gaudiAppModule.mcTCK:
            raise RuntimeError(f"{self.applicationName} step: TCK set in step, and should't be!")

        # Initialise the prodInfo object
        prodInfo = {
            "spec_version": 1,
            "application": {
                "data_pkgs": [".".join(p) for p in gaudiAppModule.extraPackages],
                "name": self.applicationName,
                "number_of_processors": self.numberOfProcessors,
                "version": gaudiAppModule.applicationVersion,
            },
            "options": {},
            "db_tags": {},
            "input": {
                "files": ["LFN:" + sid for sid in gaudiAppModule.stepInputData],
                "first_event_number": firstEventNumberGauss,
                "tck": gaudiAppModule.TCK if gaudiAppModule.TCK else gaudiAppModule.mcTCK,
                "xml_file_catalog": gaudiAppModule.poolXMLCatName,
                "xml_summary_file": gaudiAppModule.XMLSummary,
            },
            "output": {
                "prefix": gaudiAppModule.outputFilePrefix,
                "types": stepOutputTypes,
            },
        }
        if gaudiAppModule.applicationVersion.startswith("{"):
            appInfo = json.loads(gaudiAppModule.applicationVersion)
            if prodInfo["application"]["name"] != appInfo["name"]:
                raise RuntimeError(f"Application name mismatch: {prodInfo['application']['name']} != {appInfo['name']}")
            if appInfo.get("binary_tag") is not None:
                if gaudiAppModule.systemConfig != appInfo["binary_tag"]:
                    raise RuntimeError(
                        f"System config mismatch: {gaudiAppModule.systemConfig} != {appInfo['binary_tag']}"
                    )
            prodInfo["application"]["version"] = appInfo["version"]
            if appInfo.get("nightly"):
                prodInfo["application"]["nightly"] = appInfo["nightly"]

        # application
        if gaudiAppModule.systemConfig and gaudiAppModule.systemConfig.lower() != "any":
            prodInfo["application"]["binary_tag"] = gaudiAppModule.systemConfig
        prodInfo["application"]["event_timeout"] = eventTimeout

        # options
        if isinstance(commandOptions, dict):
            # This is an lbexec style application
            prodInfo["options"] = commandOptions
        else:
            prodInfo["options"]["files"] = commandOptions
            prodInfo["options"]["processing_pass"] = gaudiAppModule.processingPass
            if gaudiAppModule.optionsFormat:
                prodInfo["options"]["format"] = gaudiAppModule.optionsFormat
            if gaudiAppModule.extraOptionsLine:
                prodInfo["options"]["gaudi_extra_options"] = gaudiAppModule.extraOptionsLine

        # db_tags
        if gaudiAppModule.DDDBTag:
            if gaudiAppModule.DDDBTag.lower() == "online":
                try:
                    prodInfo["db_tags"]["dddb_tag"] = gaudiAppModule.onlineDDDBTag
                    self.log.debug("Set the online DDDB tag")
                except NameError:
                    self.log.exception("Could not find online DDDb Tag")
                    raise RuntimeError("Could not find online DDDb Tag")
            else:
                prodInfo["db_tags"]["dddb_tag"] = gaudiAppModule.DDDBTag

        if gaudiAppModule.condDBTag:
            prodInfo["db_tags"]["conddb_tag"] = (
                gaudiAppModule.onlineCondDBTag
                if gaudiAppModule.condDBTag.lower() == "online"
                else gaudiAppModule.condDBTag
            )

        if gaudiAppModule.dqTag:
            prodInfo["db_tags"]["dq_tag"] = gaudiAppModule.dqTag

        # input
        if (
            self.applicationName.lower() == "gauss"
            and gaudiAppModule.CPUe
            and gaudiAppModule.maxNumberOfEvents
            and gaudiAppModule.numberOfEvents <= 0
        ):
            # Here we set maxCPUTime to 24 hours, which seems reasonable
            prodInfo["input"]["n_of_events"] = getEventsToProduce(
                gaudiAppModule.CPUe, maxNumberOfEvents=gaudiAppModule.maxNumberOfEvents, jobMaxCPUTime=86400
            )
        else:
            prodInfo["input"]["n_of_events"] = gaudiAppModule.numberOfEvents

        if runNumberGauss:
            prodInfo["input"]["run_number"] = runNumberGauss
        if gaudiAppModule.runNumber and gaudiAppModule.runNumber not in ("Unknown", "Multiple"):
            prodInfo["input"]["run_number"] = gaudiAppModule.runNumber

        # output
        if histogram:
            prodInfo["output"]["histogram_file"] = gaudiAppModule.histoName

        with open(self.prodConfFileName, "w") as fp:
            json.dump(prodInfo, fp, indent=2)

    def run(self):
        """Invokes lb-prod-run (what you call after having setup the object)"""
        returncode, stdout, stderr = asyncio.get_event_loop().run_until_complete(self._runApp())
        if returncode != 0:
            self.log.error("lb-run or its application exited with status %d" % returncode)
            self.log.error(stderr)
            raise LHCbApplicationError(
                f"{self.applicationName} {self.applicationVersion} exited with status {returncode}"
            )

        return S_OK((returncode, stdout, stderr))

    async def _runApp(self):
        command = ["lb-prod-run", self.prodConfFileName, "--prmon", "--verbose"]
        if self.applicationName == "Gauss" and self.usePrmon:
            command = [self.prmonPath, "--json-summary", "./prmon_Gauss.json", "--"] + command
        self.log.notice("Running command", shlex.join(command))

        stdout = ""
        stderr = ""
        if gConfig.getValue("/LocalSite/RemoteExecution", False):
            remoteRunner = RemoteRunner(
                gConfig.getValue("/LocalSite/Site"),
                gConfig.getValue("/LocalSite/GridCE"),
                gConfig.getValue("/LocalSite/CEQueue"),
            )
            outputDict = remoteRunner.execute(shlex.join(command), numberOfProcessors=self.numberOfProcessors)
            if not outputDict["OK"]:
                returncode = outputDict["Errno"]
                stderr = outputDict["Message"]
            else:
                returncode, stdout, stderr = outputDict["Value"]

            self._handleRemoteOutput(stdout, self.applicationLog)
            self._handleRemoteOutput(stderr, self.stdError)
        else:
            proc = await asyncio.create_subprocess_exec(
                *command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout_fh = None
            if self.applicationLog:
                stdout_fh = open(self.applicationLog, "a")

            stderr_fh = None
            if self.stdError == self.applicationLog:
                stderr_fh = stdout_fh
            elif self.stdError:
                stderr_fh = open(self.applicationLog, "a")

            try:
                await asyncio.gather(
                    self._handleOutput(proc.stdout, stdout_fh),
                    self._handleOutput(proc.stderr, stderr_fh),
                    proc.wait(),
                )
            finally:
                if stdout_fh:
                    stdout_fh.close()
                if stderr_fh and stdout_fh != stderr_fh:
                    stderr_fh.close()
            returncode = proc.returncode
        return (returncode, stdout, stderr)

    async def _handleOutput(self, stream, fh):
        """Process the output of a current local execution"""
        while line := await stream.readline():
            line = line.decode(errors="backslashreplace")
            self._handleLine(line)
            if fh:
                fh.write(line)

    def _handleRemoteOutput(self, lines, filename):
        """Process the output of a remote execution"""
        if filename:
            with open(filename, "a") as log:
                log.write(lines)
        for line in lines.split("\n"):
            self._handleLine(line)

    def _handleLine(self, line):
        """Print a given line to the standard output if related to an event"""
        if "INFO Evt" in line or "Reading Event record" in line or "lb-run" in line:
            # These ones will appear in the std.out log too
            print(line.rstrip())



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get the GUID of a (set of) ROOT file The file can be either local, an LFN or
an xrootd URL (root:...)"""
import os

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage("\n".join([__doc__, "Usage:", f"  {Script.scriptName} [option|cfgfile] file1 [file2 ...]"]))
    Script.parseCommandLine(ignoreErrors=True)
    files = []
    for oFile in Script.getPositionalArgs():
        files += oFile.split(",")

    import DIRAC
    from DIRAC.Interfaces.API.Dirac import Dirac
    from LHCbDIRAC.Core.Utilities.File import getRootFileGUIDs
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import printDMResult

    if not files:
        Script.showHelp(exitCode=1)
    existFiles = {}
    nonExisting = []
    dirac = Dirac()

    for localFile in files:
        if os.path.exists(localFile):
            existFiles[os.path.realpath(localFile)] = localFile
        elif localFile.startswith("/lhcb"):
            res = dirac.getReplicas(localFile, active=True, preferDisk=True)
            if res["OK"] and localFile in res["Value"]["Successful"]:
                ses = list(res["Value"]["Successful"][localFile])
                for se in ses:
                    res = dirac.getAccessURL(localFile, se, protocol=["root", "xroot"])
                    if res["OK"] and localFile in res["Value"]["Successful"]:
                        existFiles[res["Value"]["Successful"][localFile]] = f"{localFile} @ {se}"
            else:
                nonExisting.append(localFile)
        elif localFile.startswith("root:"):
            existFiles[localFile] = localFile
        else:
            nonExisting.append(localFile)

    fileGUIDs = getRootFileGUIDs(list(existFiles))
    for status in ("Successful", "Failed"):
        for file in list(fileGUIDs.get("Value", {}).get(status, {})):
            fileGUIDs["Value"][status][existFiles.get(file, file)] = fileGUIDs["Value"][status].pop(file)
    if nonExisting:
        fileGUIDs["Value"]["Failed"].update(dict.fromkeys(nonExisting, "Non existing file"))
    printDMResult(fileGUIDs)

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
    LHCb class for doing consistency checks, between files in:
    - Bookkeeping
    - Transformation
    - File Catalog
"""
import asyncio
import time
import ast
import os
import copy

import DIRAC
from rich.progress import Progress, MofNCompleteColumn, TimeElapsedColumn

from DIRAC import gLogger

from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup
from DIRAC.DataManagementSystem.Client.ConsistencyInspector import ConsistencyInspector as DiracConsistencyChecks
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.Adler import compareAdler

from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


# FIXME: this is quite dirty, what should be checked is exactly what it is done
prodsWithMerge = (
    "MCSimulation",
    "MCFastSimulation",
    "DataStripping",
    "MCStripping",
    "DataSwimming",
    "WGProduction",
    "Sprucing",
)


def getFileDescendents(transID, lfns, transClient=None, dm=None, bkClient=None, descendantsDepth=None):
    """Function that returns the list of descendants from BKK

    Args:
        transID (str, int): transformationID
        lfns (str, list, dict): a string for a single lfn, a list of strings, or a dict with lfns as keys

    Returns:
        dict: a dictionary of files with descendants (lfn as key)

    Examples:
        >>> getFileDescendents(55032, ['/lhcb/LHCb/anLFN_1.dst', '/lhcb/LHCb/anLFN_2.dst',
                                       '/lhcb/LHCb/anLFN_3_NODESCENDANTS.dst'])
        {'/lhcb/LHCb/anLFN_1.dst': ['/lhcb/validation/desc_1.PIDCALIB.mdst',
                                    '/lhcb/validation/desc_1.pidcalib.root'],
        {'/lhcb/LHCb/anLFN_2.dst': ['/lhcb/validation/desc_2.PIDCALIB.mdst',
                                    '/lhcb/validation/desc_2.pidcalib.root'],

    """
    cc = ConsistencyChecks(interactive=False, transClient=transClient, dm=dm, bkClient=bkClient)
    if descendantsDepth is not None:
        cc.descendantsDepth = descendantsDepth
    else:
        cc.descendantsDepth = 1
    cc.prod = transID
    cc.fileType = []
    cc.fileTypesExcluded = Operations().getValue("DataConsistency/IgnoreDescendantsOfType", [])
    savedLevel = gLogger.getLevel()
    gLogger.setLevel("FATAL")
    result = cc.getDescendants(lfns)
    # Files with descendants
    descendants = result[0]
    # Add files with multiple descendants
    descendants.update(result[2])
    gLogger.setLevel(savedLevel)
    return descendants


class ConsistencyChecks(DiracConsistencyChecks):
    """LHCb extension to ConsistencyInspector"""

    def __init__(self, interactive=True, transClient=None, dm=None, bkClient=None, fc=None):
        """c'tor"""
        super().__init__(interactive, transClient, dm, fc)

        self.dmsHelpers = DMSHelpers()
        self.bkClient = BookkeepingClient() if bkClient is None else bkClient
        self.transClient = TransformationClient() if transClient is None else transClient

        # Base elements from which to start the consistency checks
        self._prod = 0
        self._daughtersCache = None
        self._bkQuery = None
        self._fileType = []
        self._fileTypesExcluded = []
        self._lfns = []
        self._status = None
        self._seList = []
        self._verbose = False
        self.noFC = False
        self.directories = []
        self.descendantsDepth = 10
        self.ancestorsDepth = 10

        # Results of the checks
        self.existLFNsBKRepNo = {}
        self.absentLFNsBKRepNo = []
        self.existLFNsBKRepYes = []
        self.absentLFNsBKRepYes = []
        self.existLFNsNotInBK = []
        self.absentLFNsNotInBK = []
        self.filesInBKNotInTS = []

        self.inBKNotInFC = []
        self.inFCNotInBK = []

        self.removedFiles = []
        self.inFailover = []

        self.absentLFNsInFC = []
        self.inSEbutNotInFC = {}
        self.notRegisteredAtSE = {}
        self.existLFNsNoSE = {}
        self.existLFNsBadReplicas = {}
        self.existLFNsBadFiles = {}
        self.existLFNsNotExisting = {}
        self.commonAncestors = {}
        self.multipleDescendants = {}
        self.ancestors = {}

        self._verbose = False
        self._seList = None

    def __logVerbose(self, msg, msg1=""):
        if self._verbose:
            newMsg = "[ConsistencyChecks] " + f"[{str(self.prod)}] " if self.prod else ""
            # Add that prefix to all lines of the message
            newMsg1 = msg1.replace("\n", "\n" + newMsg)
            newMsg += msg.replace("\n", "\n" + newMsg)
            gLogger.notice(newMsg, newMsg1)
        else:
            gLogger.verbose(msg, msg1)

    ################################################################################

    def __getLFNsFromBK(self, checkAll=False):
        """
        Get list of LFNs with gotReplica No and Yes
        """
        lfnsReplicaNo, lfnsReplicaYes = (0, 0)
        if self.lfns:
            lfnsNotInBK, lfnsReplicaNo, lfnsReplicaYes = self._getBKMetadata(self.lfns)
            lfnsReplicaNo = list(lfnsReplicaNo) + lfnsNotInBK
        else:
            bkQuery = self.__getBKQuery()
            gLogger.notice(f"Getting files for BK query {str(bkQuery)}...")
            if checkAll:
                lfnsReplicaNo = self._getBKFiles(bkQuery, "No")
            lfnsReplicaYes = self._getBKFiles(bkQuery, "Yes")
        return lfnsReplicaNo, lfnsReplicaYes

    def checkBK2FC(self, checkAll):
        """Starting from the BK, check if the FileCatalog has consistent information (BK -> FileCatalog)

        Works either when the bkQuery is free, or when it is made using a transformation ID
        """
        lfnsReplicaNo, lfnsReplicaYes = self.__getLFNsFromBK(checkAll)

        if self.lfns:
            self.__logVerbose("Checking the File Catalog for those files with BK ReplicaFlag = No")
            self.existLFNsBKRepNo, self.absentLFNsBKRepNo = self.getReplicasPresence(lfnsReplicaNo)
            self.__logVerbose("Checking the File Catalog for those files with BK ReplicaFlag = Yes")
            self.existLFNsBKRepYes, self.absentLFNsBKRepYes = self.getReplicasPresence(lfnsReplicaYes)
        elif self.transType not in prodsWithMerge:
            # Merging and Reconstruction
            # In principle few files without replica flag, check them in FC
            if checkAll:
                self.__logVerbose("Checking the File Catalog for those files with BK ReplicaFlag = No")
                self.existLFNsBKRepNo, self.absentLFNsBKRepNo = self.getReplicasPresence(lfnsReplicaNo)
            self.__logVerbose("Checking the File Catalog for those files with BK ReplicaFlag = Yes")
            self.existLFNsBKRepYes, self.absentLFNsBKRepYes = self.getReplicasPresenceFromDirectoryScan(lfnsReplicaYes)

        else:
            # prodsWithMerge
            # In principle most files have no replica flag, start from the File Catalog files with replicas
            if checkAll:
                self.__logVerbose("Checking the File Catalog for those files with BK ReplicaFlag = No")
                self.existLFNsBKRepNo, self.absentLFNsNotInBK = self.getReplicasPresenceFromDirectoryScan(lfnsReplicaNo)
            self.__logVerbose("Checking the File Catalog for those files with BK ReplicaFlag = Yes")
            self.existLFNsBKRepYes, self.absentLFNsBKRepYes = self.getReplicasPresence(lfnsReplicaYes)

        if checkAll and self.existLFNsBKRepNo:
            msg = "%d files have ReplicaFlag = No, but %d are in the FC" % (
                len(lfnsReplicaNo),
                len(self.existLFNsBKRepNo),
            )
            if self.transType:
                msg = f"For prod {self.prod} of type {self.transType}, " + msg
            gLogger.info(msg)

        if self.absentLFNsBKRepYes:
            msg = "%d files have ReplicaFlag = Yes, but %d are not in the FC" % (
                len(lfnsReplicaYes),
                len(self.absentLFNsBKRepYes),
            )
            if self.transType:
                msg = f"For prod {self.prod} of type {self.transType}, " + msg
            gLogger.info(msg)

        ################################################################################

    def _getBKFiles(self, bkQuery, replicaFlag="Yes"):
        """Helper function - get files from BK, first constructing the bkQuery"""
        visibility = bkQuery.isVisible()
        if self.transType:
            visibility = "Yes" if self.transType not in prodsWithMerge else "All"
        bkQuery.setVisible("All")
        bkQueryRes = BKQuery(bkQuery, visible=visibility)
        bkQueryRes.setOption("ReplicaFlag", replicaFlag)
        startTime = time.time()
        lfnsRes = bkQueryRes.getLFNs(printOutput=False)
        if not lfnsRes:
            gLogger.info(f"No files found with replica flag = {replicaFlag}")
        else:
            gLogger.info(
                "Found %d files with replica flag = %s (%.1f seconds)"
                % (len(lfnsRes), replicaFlag, time.time() - startTime)
            )

        return lfnsRes

    def __getBKQuery(self, fromTS=False):
        """get the bkQuery to be used"""
        bkQuery = None
        if fromTS:
            res = self.transClient.getBookkeepingQuery(self.prod)
            if not res["OK"]:
                raise ValueError(res["Message"])
            bkQuery = BKQuery(res["Value"])
        else:
            if self.bkQuery:
                bkQuery = self.bkQuery
            if self.prod:
                # If production is specificed, the visibility flag is ignored
                if not self.bkQuery:
                    bkQuery = BKQuery(prods=self.prod, fileTypes=self.fileType, visible=None)
                else:
                    bkQuery = BKQuery(self.bkQuery.setOption("Production", self.prod), visible=None)
            if not bkQuery:
                raise ValueError("Need to specify either the bkQuery or a production id")

        return bkQuery

    ################################################################################

    def getReplicasPresence(self, lfns, ignoreFailover=False, typeStr="files"):
        """get the replicas using the standard DataManager.getReplicas()"""
        present = set()
        notPresent = set()
        if not lfns:
            return [], []
        lfns = set(lfns)

        chunkSize = 500
        if not ignoreFailover:
            progressBar = ProgressBar(
                len(lfns),
                title=f"Checking replicas for {len(lfns)} {typeStr}",
                chunk=chunkSize,
                interactive=self.interactive,
            )
        else:
            progressBar = None
        for chunk in breakListIntoChunks(lfns, chunkSize):
            if progressBar:
                progressBar.loop()
            for _ in range(1, 10):
                res = self.dataManager.getReplicas(chunk, getUrl=False)
                if res["OK"]:
                    success = res["Value"]["Successful"]
                    if ignoreFailover:
                        present.update(
                            lfn
                            for lfn, seList in success.items()
                            for se in seList
                            if not self.dmsHelpers.isSEFailover(se)
                        )
                    else:
                        present.update(success)
                    self.cachedReplicas.update(success)
                    notPresent.update(res["Value"]["Failed"])
                    break
                time.sleep(0.1)
        if progressBar:
            progressBar.endLoop(f"found {len(present)} files with replicas and {len(notPresent)} without")

        if notPresent:
            self.__logVerbose("Files without replicas:", "\n".join([""] + sorted(notPresent)))
        return list(present), list(notPresent)

    ################################################################################

    def getReplicasPresenceFromDirectoryScan(self, lfns, typeStr="files"):
        """Get replicas scanning the directories. Might be faster."""

        dirs = {}
        present = []
        notPresent = []
        compare = True

        for lfn in lfns:
            dirN = os.path.dirname(lfn)
            if lfn == dirN + "/":
                compare = False
            dirs.setdefault(dirN, []).append(lfn)

        if compare:
            title = f"Checking File Catalog for {len(lfns)} {typeStr} from {len(dirs)} directories "
        else:
            title = f"Getting files from {len(dirs)} directories "
        progressBar = ProgressBar(len(dirs), title=title)

        for dirN in sorted(dirs):
            progressBar.loop()
            startTime = time.time()
            lfnsFound = self._getFilesFromDirectoryScan(dirN)
            self.__logVerbose(f"Obtained {len(lfnsFound)} files in {time.time() - startTime:.1f} seconds")
            if compare:
                pr, notPr = self.__compareLFNLists(dirs[dirN], lfnsFound)
                notPresent += notPr
                present += pr
            else:
                present += lfnsFound

        progressBar.endLoop(f"found {len(present)} files with replicas and {len(notPresent)} without")
        return present, notPresent

    ################################################################################

    def __compareLFNLists(self, lfns, lfnsFound):
        """return files in both lists and files in lfns and not in lfnsFound"""
        present = []
        notPresent = lfns
        startTime = time.time()
        self.__logVerbose(f"Comparing list of {len(lfns)} LFNs with second list of {len(lfnsFound)}")
        if lfnsFound:
            # print sorted( lfns )
            # print sorted( lfnsFound )
            setLfns = set(lfns)
            setLfnsFound = set(lfnsFound)
            present = list(setLfns & setLfnsFound)
            notPresent = list(setLfns - setLfnsFound)
            # print sorted( present )
            # print sorted( notPresent )
        self.__logVerbose(f"End of comparison: {time.time() - startTime:.1f} seconds")
        return present, notPresent

    def _getFilesFromDirectoryScan(self, dirs):
        """calls dm.getFilesFromDirectory"""

        level = gLogger.getLevel()
        gLogger.setLevel("FATAL")
        res = self.dataManager.getFilesFromDirectory(dirs)
        gLogger.setLevel(level)
        if not res["OK"]:
            if "No such file or directory" not in res["Message"]:
                gLogger.error(f"Error getting files from directories {dirs}:", res["Message"])
            return []
        if res["Value"]:
            lfnsFound = res["Value"]
        else:
            lfnsFound = []

        return lfnsFound

    ################################################################################

    def checkTS2BK(self):
        """Check if lfns has descendants (TransformationFiles -> BK)"""
        if not self.prod:
            raise ValueError("You need a transformationID")

        gLogger.notice("Getting files from the TransformationSystem...")
        startTime = time.time()
        processedLFNs, nonProcessedLFNs, statuses = self._getTSFiles()
        gLogger.notice(
            "Found %d processed files and %d non processed%s files (%.1f seconds)"
            % (
                len(processedLFNs),
                len(nonProcessedLFNs),
                " (%s)" % ",".join(statuses) if statuses else "",
                (time.time() - startTime),
            )
        )

        res = self.getDescendants(processedLFNs, status="processed")
        self.prcdWithDesc = res[0]
        self.prcdWithoutDesc = res[1]
        self.prcdWithMultDesc = res[2]
        self.descForPrcdLFNs = res[3]
        self.inFCNotInBK = res[4]
        self.inBKNotInFC = res[5]
        self.removedFiles = res[6]
        self.inFailover = res[7]

        res = self.getDescendants(nonProcessedLFNs, status="non-processed")
        self.nonPrcdWithDesc = res[0]
        self.nonPrcdWithoutDesc = res[1]
        self.nonPrcdWithMultDesc = res[2]
        self.descForNonPrcdLFNs = res[3]
        self.inFCNotInBK += res[4]
        self.inBKNotInFC += res[5]
        self.inFailover += res[7]

        prTuple = (self.prod, self.transType, len(processedLFNs))
        if self.prcdWithoutDesc:
            self.__logVerbose(
                "For prod %s of type %s, %d files are processed," % prTuple,
                f"and {len(self.prcdWithoutDesc)} of those do not have descendants",
            )

        if self.prcdWithMultDesc:
            self.__logVerbose(
                "For prod %s of type %s, %d files are processed," % prTuple,
                f"and {len(self.prcdWithMultDesc)} of those have multiple descendants: ",
            )

        prTuple = (self.prod, self.transType, len(nonProcessedLFNs))
        if self.nonPrcdWithDesc:
            self.__logVerbose(
                "For prod %s of type %s, %d files are not processed," % prTuple,
                f"but {len(self.nonPrcdWithDesc)} of those have descendants",
            )

        if self.nonPrcdWithMultDesc:
            self.__logVerbose(
                "For prod %s of type %s, %d files are not processed," % prTuple,
                f"but {len(self.nonPrcdWithMultDesc)} of those have multiple descendants: ",
            )

    ################################################################################

    def checkAncestors(self):
        """Check if a set of files don't share a common ancestor"""
        if self.lfns:
            files = self.lfns
            bkQuery = None
            fileType = []
        else:
            bkQuery = self.__getBKQuery()
            gLogger.notice(f"Getting files for BK query: {bkQuery}")
            fileType = bkQuery.getFileTypeList()
            files = self._getBKFiles(bkQuery)

        if len(fileType) == 1:
            fileTypes = {fileType[0]: set(files)}
            getFileType = False
        else:
            fileTypes = {}
            getFileType = True

        chunkSize = 500
        ancestors = {}
        listAncestors = []
        progressBar = ProgressBar(
            len(files),
            title=f"Getting ancestors for {len(files)} files",
            chunk=chunkSize,
            interactive=self.interactive,
        )
        for lfnChunk in breakListIntoChunks(files, chunkSize):
            progressBar.loop()
            if getFileType:
                res = self.bkClient.getFileMetadata(lfnChunk)
                if not res["OK"]:
                    gLogger.fatal("Error getting files metadata", res["Message"])
                    DIRAC.exit(2)
                for lfn, metadata in res["Value"]["Successful"].items():
                    fileType = metadata.get("FileType")
                    if fileType is None:
                        gLogger.notice(f"File type unavailable for {lfn}", str(metadata))
                    fileTypes.setdefault(fileType, set()).add(lfn)
            res = self.bkClient.getFileAncestors(
                lfnChunk, depth=self.ancestorsDepth, replica=(self.ancestorsDepth == 10)
            )
            if not res["OK"]:
                gLogger.fatal("Error getting file ancestors", res["Message"])
                DIRAC.exit(2)
            for lfn, anc in res["Value"]["Successful"].items():
                ancestors[lfn] = [ancDict["FileName"] for ancDict in anc]
                if not getFileType:
                    listAncestors += ancestors[lfn]
        progressBar.endLoop()

        self.ancestors = ancestors
        self.commonAncestors = {}
        self.multipleDescendants = {}
        if not getFileType and len(listAncestors) == len(set(listAncestors)):
            gLogger.notice(f"Found {len(listAncestors)} ancestors, no common one")
            return

        gLogger.notice(f"Found files with {len(fileTypes)} file types")
        for fileType, lfnList in fileTypes.items():
            lfns = lfnList & set(ancestors)
            gLogger.notice(f"File type {fileType}, checking {len(lfns)} files")
            listAncestors = []
            for lfn in lfns:
                listAncestors += ancestors[lfn]
            setAncestors = set(listAncestors)
            if len(listAncestors) == len(setAncestors):
                gLogger.notice(f"Found {len(listAncestors)} ancestors for file type {fileType}, no common one")
                continue
            # There are common ancestors
            descendants = {}
            # Reverse the list of ancestors
            for lfn in lfns:
                for anc in ancestors[lfn]:
                    descendants.setdefault(anc, []).append(lfn)
            # Check if ancestor has more than one descendant
            for anc in sorted(descendants):
                if len(descendants[anc]) > 1:
                    desc = sorted(descendants[anc])
                    gLogger.notice(f"For ancestor {anc}, found {len(desc)} descendants: {desc}")
                    self.multipleDescendants[anc] = desc
                    self.commonAncestors.setdefault(",".join(sorted(desc)), []).append(anc)

    ################################################################################

    def _getTSFiles(self):
        """Helper function - get files from the TS"""
        selectDict = {"TransformationID": self.prod}
        if self._lfns:
            selectDict["LFN"] = self._lfns
        elif self._status:
            selectDict["Status"] = self._status
        elif self.runStatus and self.fromProd:
            res = self.transClient.getTransformationRuns({"TransformationID": self.fromProd, "Status": self.runStatus})
            if not res["OK"]:
                gLogger.error("Failed to get runs for transformation %d" % self.prod)
            else:
                if res["Value"]:
                    self.runsList.extend(
                        run["RunNumber"] for run in res["Value"] if run["RunNumber"] not in self.runsList
                    )
                    gLogger.notice(
                        f"{len(res['Value'])} runs selected:", ",".join(str(run["RunNumber"]) for run in res["Value"])
                    )
                elif not self.runsList:
                    gLogger.notice("No runs selected, check completed")
                    DIRAC.exit(0)
        if not self._lfns and self.runsList:
            selectDict["RunNumber"] = self.runsList

        res = self.transClient.getTransformation(self.prod)
        if not res["OK"]:
            gLogger.error(f"Failed to find transformation {self.prod}")
            return [], [], []
        status = res["Value"]["Status"]
        if status not in ("Active", "Stopped", "Completed", "Idle"):
            gLogger.notice(f"Transformation {self.prod} in status {status}, will not check if files are processed")
            processedLFNs = []
            nonProcessedLFNs = []
            nonProcessedStatuses = []
            if self._lfns:
                processedLFNs = self._lfns
        else:
            res = self.transClient.getTransformationFiles(selectDict, columns=["LFN", "Status"])
            if not res["OK"]:
                gLogger.error("Failed to get files for transformation %d" % self.prod, res["Message"])
                return [], [], []
            processedLFNs = [item["LFN"] for item in res["Value"] if item["Status"] == "Processed"]
            nonProcessedLFNs = [item["LFN"] for item in res["Value"] if item["Status"] != "Processed"]
            nonProcessedStatuses = list({item["Status"] for item in res["Value"] if item["Status"] != "Processed"})

        return processedLFNs, nonProcessedLFNs, nonProcessedStatuses

    ################################################################################

    async def __getDaughtersInfoInner(self):
        PROGRESS_COLUMNS = [MofNCompleteColumn(), *Progress.get_default_columns(), TimeElapsedColumn()]

        async def do_work(prod_id, run_number):
            loop = asyncio.get_running_loop()
            return returnValueOrRaise(
                await loop.run_in_executor(None, BookkeepingClient().getInputOutputFilesForProd, prod_id, run_number)
            )

        runs = returnValueOrRaise(self.bkClient.getListOfRunsInProd(self.prod))

        result = {}
        with Progress(*PROGRESS_COLUMNS) as progress:
            task = progress.add_task("Getting file decendents...", total=len(runs))
            for coro in asyncio.as_completed(map(lambda x: do_work(self.prod, x), runs)):
                progress.advance(task)
                coro_result = await coro
                aname_idx = coro_result["ParameterNames"].index("AncestorLFN")
                dname_idx = coro_result["ParameterNames"].index("DescendantLFN")
                gotreplica_idx = coro_result["ParameterNames"].index("GotReplica")
                filetype_idx = coro_result["ParameterNames"].index("FileType")
                for record in coro_result["Records"]:
                    result.setdefault(record[aname_idx], {})[record[dname_idx]] = {
                        "GotReplica": record[gotreplica_idx],
                        "FileType": record[filetype_idx],
                    }
        return result

    def __getDaughtersInfoOld(
        self, lfns, status, filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants
    ):
        """Get BK information about daughers of lfns"""
        chunkSize = 20
        lfns = set(lfns)
        progressBar = ProgressBar(
            len(lfns),
            title="Now getting all daughters for %d %s mothers in production %d (depth %d)"
            % (len(lfns), status, self.prod, self.descendantsDepth),
            chunk=chunkSize,
            interactive=self.interactive,
        )
        daughtersBKInfo = {}
        for lfnChunk in breakListIntoChunks(lfns, chunkSize):
            progressBar.loop()
            while True:
                resChunk = self.bkClient.getFileDescendents(
                    lfnChunk, depth=self.descendantsDepth, production=self.prod, checkreplica=False
                )
                # If error, global or for some files, retry
                if resChunk["OK"] and not resChunk["Value"]["Failed"]:
                    # Key is ancestor, value is metadata dictionary of daughters
                    descDict = self._selectByFileType(resChunk["Value"]["WithMetadata"])
                    # Do the daughters have a replica flag in BK? Store file type as well... Key is daughter
                    daughtersBKInfo.update(
                        {
                            lfn: (desc[lfn]["GotReplica"] == "Yes", desc[lfn]["FileType"])
                            for desc in descDict.values()
                            for lfn in desc
                        }
                    )
                    # Count the daughters per file type (key is ancestor)
                    ftCount = self._getFileTypesCount(descDict)
                    for lfn in lfnChunk:
                        # Check if file has a daughter and how many per file type
                        if lfn in descDict:
                            # Assign the daughters list to the initial LFN
                            filesWithDescendants[lfn] = list(descDict[lfn])
                            # Is there a file type with more than one daughter of a given file type?
                            multi = {ft: ftc for ft, ftc in ftCount[lfn].items() if ftc > 1}
                            if multi:
                                filesWithMultipleDescendants[lfn] = multi
                        else:
                            # No daughter, easy case!
                            filesWithoutDescendants[lfn] = None
                    break
                progressBar.comment(f"Error getting daughters for {len(lfnChunk)} files, retry", resChunk["Message"])
        prStr = ""
        if filesWithDescendants:
            nb = sum(len(desc) for desc in filesWithDescendants.values())
            prStr += "found %d descendants (%d unique) for %d files" % (
                nb,
                len(daughtersBKInfo),
                len(filesWithDescendants),
            )
        if filesWithoutDescendants:
            if not prStr:
                prStr = "found"
            else:
                prStr += " and"
            prStr += f" no descendants for {len(filesWithoutDescendants)} files"
        progressBar.endLoop(message=prStr)
        return daughtersBKInfo

    def __getDaughtersInfo(
        self, lfns, status, filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants
    ):
        if len(lfns) < 200:
            return self.__getDaughtersInfoOld(
                lfns, status, filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants
            )
        if self._daughtersCache is None:
            self._daughtersCache = asyncio.run(self.__getDaughtersInfoInner())

        # Key is ancestor, value is metadata dictionary of daughters
        descDict = self._selectByFileType({k: self._daughtersCache[k] for k in lfns if k in self._daughtersCache})
        # Do the daughters have a replica flag in BK? Store file type as well... Key is daughter
        daughtersBKInfo = {
            lfn: (desc[lfn]["GotReplica"] == "Yes", desc[lfn]["FileType"]) for desc in descDict.values() for lfn in desc
        }

        # Count the daughters per file type (key is ancestor)
        ftCount = self._getFileTypesCount(descDict)
        for lfn in lfns:
            # Check if file has a daughter and how many per file type
            if lfn in descDict:
                # Assign the daughters list to the initial LFN
                filesWithDescendants[lfn] = list(descDict[lfn])
                # Is there a file type with more than one daughter of a given file type?
                multi = {ft: ftc for ft, ftc in ftCount[lfn].items() if ftc > 1}
                if multi:
                    filesWithMultipleDescendants[lfn] = multi
            else:
                # No daughter, easy case!
                filesWithoutDescendants[lfn] = None

        prStr = []
        if filesWithDescendants:
            nb = sum(len(desc) for desc in filesWithDescendants.values())
            prStr.append(
                f"found {nb} descendants ({len(daughtersBKInfo)} unique) for {len(filesWithDescendants)} files"
            )
        if filesWithoutDescendants:
            prStr.append(f"no descendants for {len(filesWithoutDescendants)} files")
        if prStr:
            gLogger.info("found " + " and ".join(prStr))

        return daughtersBKInfo

    async def _getBulkDescendants(self):
        PROGRESS_COLUMNS = [MofNCompleteColumn(), *Progress.get_default_columns(), TimeElapsedColumn()]

        async def do_work(prod_id, run_number):
            loop = asyncio.get_running_loop()
            return returnValueOrRaise(
                await loop.run_in_executor(None, BookkeepingClient().getOutputDescendantsForProd, prod_id, run_number)
            )

        runs = returnValueOrRaise(self.bkClient.getListOfRunsInProd(self.prod))

        result = {}
        with Progress(*PROGRESS_COLUMNS) as progress:
            task = progress.add_task("Getting file decendents...", total=len(runs))
            for coro in asyncio.as_completed(map(lambda x: do_work(self.prod, x), runs)):
                progress.advance(task)
                coro_result = await coro
                aname_idx = coro_result["ParameterNames"].index("AncestorLFN")
                dname_idx = coro_result["ParameterNames"].index("DescendantLFN")
                gotreplica_idx = coro_result["ParameterNames"].index("GotReplica")
                filetype_idx = coro_result["ParameterNames"].index("FileType")
                for record in coro_result["Records"]:
                    if record[filetype_idx] == "LOG":
                        continue
                    if record[gotreplica_idx] != "Yes":
                        continue
                    result.setdefault(record[aname_idx], {})[record[dname_idx]] = {
                        "GotReplica": record[gotreplica_idx],
                        "FileType": record[filetype_idx],
                    }
        return result

    def getDescendants(self, lfns, status=""):
        """get the descendants of a list of LFN (for the production)

        Args:
            lfns (str, list, dict): a string for a single lfn, a list of strings, or a dict with lfns as keys
        """
        if isinstance(lfns, str):
            lfns = [lfns]
        elif isinstance(lfns, dict):
            lfns = list(lfns)
        filesWithDescendants = {}
        filesWithoutDescendants = {}
        filesWithMultipleDescendants = {}
        fileTypesExcluded = Operations().getValue("DataConsistency/IgnoreDescendantsOfType", [])
        if self.fileType:
            fileTypesExcluded = list(set(fileTypesExcluded) - set(self.fileType))
        inFCNotInBK = []
        inBKNotInFC = []
        allDaughters = []
        removedFiles = []
        inFailover = []
        if not lfns:
            return (
                filesWithDescendants,
                filesWithoutDescendants,
                filesWithMultipleDescendants,
                allDaughters,
                inFCNotInBK,
                inBKNotInFC,
                removedFiles,
                inFailover,
            )

        daughtersBKInfo = self.__getDaughtersInfo(
            lfns, status, filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants
        )
        for daughter in list(daughtersBKInfo):
            # Ignore the daughters that have a type to ignore
            if daughtersBKInfo[daughter][1] in fileTypesExcluded:
                daughtersBKInfo.pop(daughter)

        # This is the list of all daughters, sets will contain unique entries
        setAllDaughters = set(daughtersBKInfo)
        allDaughters = list(setAllDaughters)
        inBK = {lfn for lfn in setAllDaughters if daughtersBKInfo[lfn][0]}
        setRealDaughters = set()
        # Now check whether these daughters files have replicas or have descendants that have replicas
        chunkSize = 500
        if filesWithDescendants:
            # First check in FC the presence of daughters
            if not self.noFC:
                present, notPresent = (
                    self.getReplicasPresenceFromDirectoryScan(allDaughters, typeStr="daughters")
                    if len(allDaughters) > 10 * chunkSize and len(inBK) < len(allDaughters) / 2
                    else self.getReplicasPresence(allDaughters, typeStr="daughters")
                )
                setPresent = set(present)
                setNotPresent = set(notPresent)
            else:
                setPresent = inBK
                setNotPresent = setAllDaughters - inBK

            setRealDaughters = setPresent
            # Now check consistency between BK and FC for daughters
            inBKNotInFC = list(inBK & setNotPresent)
            inFCNotInBK = list(setPresent - inBK)

            # Now check whether the daughters without replica have a descendant
            if setNotPresent:
                if len(setNotPresent) > 10_000:
                    notPresentDescendants = asyncio.run(self._getBulkDescendants())
                else:
                    progressBar = ProgressBar(
                        len(setNotPresent),
                        title=f"Now getting descendants from {len(setNotPresent)} daughters without replicas",
                        chunk=chunkSize,
                        interactive=self.interactive,
                    )
                    # Get existing descendants of notPresent daughters
                    notPresentDescendants = {}
                    for lfnChunk in breakListIntoChunks(setNotPresent, chunkSize):
                        progressBar.loop()
                        while True:
                            res = self.bkClient.getFileDescendents(lfnChunk, depth=99, checkreplica=True)
                            if res["OK"]:
                                # Exclude ignored file types, but select any other file type, key is daughters
                                notPresentDescendants.update(res["Value"]["WithMetadata"])
                                break
                            progressBar.comment(
                                f"Error getting descendants for {len(lfnChunk)} files, retry", res["Message"]
                            )
                        progressBar.endLoop(message=f"found descendants for {len(notPresentDescendants)} daughters")
                uniqueDescendants = {lfn for desc in notPresentDescendants.values() for lfn in desc}
                gLogger.notice(f"Found {len(uniqueDescendants)} unique descendants")
                # Check if descendants have a replica in the FC
                setDaughtersWithDesc = set()
                if uniqueDescendants:
                    _, notPresent = self.getReplicasPresence(uniqueDescendants)
                    inBKNotInFC += notPresent
                    # Remove descendants that are not in FC, and if no descendants remove ancestor as well
                    for anc in list(notPresentDescendants):
                        for desc in list(notPresentDescendants[anc]):
                            if desc in notPresent:
                                notPresentDescendants[anc].pop(desc)
                        if not notPresentDescendants[anc]:
                            notPresentDescendants.pop(anc)
                    if notPresentDescendants:
                        setDaughtersWithDesc = set(
                            self._selectByFileType(
                                notPresentDescendants, fileTypes=[""], fileTypesExcluded=fileTypesExcluded
                            )
                        )

                progressBar = ProgressBar(
                    len(filesWithDescendants),
                    title="Now establishing final list of existing descendants for %d mothers"
                    % len(filesWithDescendants),
                    step=chunkSize,
                    interactive=self.interactive,
                )
                for lfn in set(filesWithDescendants):
                    setDaughters = set(filesWithDescendants[lfn])
                    progressBar.loop()
                    # If all daughters are present, all is easy...
                    daughtersNotPresent = setDaughters & setNotPresent
                    if not daughtersNotPresent:
                        continue
                    self.__logVerbose("Mother file:", lfn)
                    self.__logVerbose("Daughters:\n", "\n".join(sorted(filesWithDescendants[lfn])))
                    self.__logVerbose("Not present daughters:\n", "\n".join(sorted(daughtersNotPresent)))
                    notPresentWithDesc = daughtersNotPresent & setDaughtersWithDesc
                    if notPresentWithDesc:
                        self.__logVerbose(" but with descendants:\n", "\n".join(sorted(notPresentWithDesc)))
                    else:
                        self.__logVerbose(" none of them has descendants")
                    # Some daughters may have a replica though, take them into account
                    daughtersWithReplica = setDaughters & setPresent
                    # and add those without a replica but that have  a descendant with replica
                    realDaughters = daughtersWithReplica | notPresentWithDesc
                    if realDaughters:
                        self.__logVerbose("Real Daughters:\n", "\n".join(sorted(realDaughters)))
                    else:
                        self.__logVerbose("No real daughters found")
                    # descToCheck: dictionary with key = daughter and value = dictionary of file type counts
                    daughtersDict = {
                        daughter: {daughter: {"FileType": daughtersBKInfo[daughter][1]}} for daughter in realDaughters
                    }
                    descToCheck = self._getFileTypesCount(daughtersDict)

                    # Update the result dictionaries according to the final set of descendants
                    if not descToCheck:
                        # Mother has no descendant
                        self.__logVerbose(f"{lfn} has no real descendants")
                        filesWithMultipleDescendants.pop(lfn, None)
                        filesWithDescendants.pop(lfn, None)
                        filesWithoutDescendants[lfn] = None
                    else:
                        self.__logVerbose("Descendants to check:\n", "\n".join(sorted(descToCheck)))
                        filesWithDescendants[lfn] = sorted(realDaughters)
                        setRealDaughters.update(realDaughters)
                        # Count the descendants by file type
                        ftCount = {}
                        for counts in descToCheck.values():
                            for ft in counts:
                                ftCount[ft] = ftCount.setdefault(ft, 0) + counts.get(ft, 0)
                        multi = {ft: ftc for ft, ftc in ftCount.items() if ftc > 1}
                        # Mother has at least one real descendant
                        # Now check whether there are more than one descendant of the same file type
                        if not multi:
                            filesWithMultipleDescendants.pop(lfn, None)
                            prStr = "single"
                        else:
                            filesWithMultipleDescendants[lfn] = multi
                            prStr = "multiple"
                        self.__logVerbose(f"Found {prStr} descendants")
                progressBar.endLoop(message=f"found {len(filesWithDescendants)} files with existing descendants")
        # print 'Final multiple descendants', filesWithMultipleDescendants

        # File files without descendants don't exist, not important
        if filesWithoutDescendants:
            present, removedFiles = self.getReplicasPresence(list(filesWithoutDescendants), ignoreFailover=True)
            filesWithoutDescendants = dict.fromkeys(present)
        else:
            removedFiles = []

        # Remove files with multiple descendants from files with descendants
        for lfn in filesWithMultipleDescendants:
            filesWithDescendants.pop(lfn, None)
        # For files in FC and not in BK, ignore if they are not active
        if inFCNotInBK:
            progressBar = ProgressBar(
                len(inFCNotInBK),
                title=f"Checking Failover for {len(inFCNotInBK)} descendants found in FC and not in BK",
                step=1,
                interactive=self.interactive,
            )
            notInFailover, _notFound = self.getReplicasPresence(inFCNotInBK, ignoreFailover=True)
            inFailover = list(set(inFCNotInBK) - set(notInFailover))
            progressBar.endLoop(message=f"found {len(inFailover)} in Failover")
            inFCNotInBK = notInFailover
        return (
            filesWithDescendants,
            filesWithoutDescendants,
            filesWithMultipleDescendants,
            list(setRealDaughters),
            inFCNotInBK,
            inBKNotInFC,
            removedFiles,
            inFailover,
        )

    ################################################################################

    def _selectByFileType(self, lfnDict, fileTypes=None, fileTypesExcluded=None, lfns=set()):
        """Select only those files from the values of lfnDict that have a certain type"""
        if not lfnDict:
            return {}
        if not fileTypes:
            fileTypes = self.fileType
        if not fileTypesExcluded:
            fileTypesExcluded = self.fileTypesExcluded
        else:
            fileTypesExcluded += [ft for ft in self.fileTypesExcluded if ft not in fileTypesExcluded]
        if fileTypes == [""]:
            fileTypes = []
        if fileTypes:
            fileTypesExcluded = list(set(fileTypesExcluded) - set(fileTypes))
        # Select only files with the specified file type constraints
        ancDict = {
            ancestor: {
                k: v
                for k, v in lfnDict[ancestor].items()
                if (not fileTypes or v["FileType"] in fileTypes) and (v["FileType"] not in fileTypesExcluded)
            }
            for ancestor in lfnDict
        }
        # Remove files which no longer have descendants
        return {k: v for k, v in ancDict.items() if v}

    @staticmethod
    def _getFileTypesCount(lfnDict):
        """return file types count"""
        ftDict = {}
        for ancestor in lfnDict:
            tDict = {}
            for desc in lfnDict[ancestor]:
                ft = lfnDict[ancestor][desc]["FileType"]
                tDict[ft] = tDict.setdefault(ft, 0) + 1
            ftDict[ancestor] = tDict

        return ftDict

    def __getLFNsFromFC(self):
        if not self.lfns:
            directories = []
            for dirName in self.__getDirectories():
                if not dirName.endswith("/"):
                    dirName += "/"
                directories.append(dirName)
            present, notPresent = self.getReplicasPresenceFromDirectoryScan(directories)
        else:
            present, notPresent = self.getReplicasPresence(self.lfns)
        return present, notPresent

    def _checkFilesInSE(self, notPresent, seList):
        foundInSE = {}
        for se in seList:
            seObj = StorageElement(se)
            res = seObj.exists(notPresent)
            if not res["OK"]:
                gLogger.error("Error checking file in SE", res["Message"])
            else:
                for lfn, ex in res["Value"]["Successful"].items():
                    if ex:
                        foundInSE.setdefault(lfn, []).append(se)
        return foundInSE

    def checkFC2BK(self, bkCheck=True):
        """check that files present in the FC are also in the BK"""
        present, notPresent = self.__getLFNsFromFC()
        foundInSE = {}
        if not self.lfns:
            prStr = " are in the FC but"
        else:
            if notPresent and self._seList:
                gLogger.notice(f"Found {len(notPresent)} files not in FC, check if they are in specified SEs")
                foundInSE = self._checkFilesInSE(notPresent, self._seList)
                if foundInSE:
                    self.inSEbutNotInFC = foundInSE
            elif not present:
                if bkCheck:
                    gLogger.notice("No files are in the FC, no check in the BK. Use dirac-dms-check-bkk2fc instead")
                return
            prStr = ""

        if bkCheck and (present or foundInSE):
            res = self._getBKMetadata(present + list(foundInSE))
            self.existLFNsNotInBK = res[0]
            self.existLFNsBKRepNo = res[1]
            self.existLFNsBKRepYes = res[2]
            msg = ""
            if self.transType:
                msg = f"For prod {self.prod} of type {self.transType}, "
            if self.existLFNsBKRepNo:
                gLogger.warn(f"{msg} {len(self.existLFNsBKRepNo)} files{prStr} have replica = NO in BK")
            if self.existLFNsNotInBK:
                gLogger.warn(f"{msg} {len(self.existLFNsNotInBK)} files{prStr} not in BK")
        else:
            self.existLFNsBKRepYes = present

        ########################################################################

    def __getDirectories(self):
        """get the directories where to look into (they are either given, or taken from the transformation ID"""
        if self.directories:
            directories = []
            printout = False
            for directory in self.directories:
                if not directory.endswith("..."):
                    directories.append(directory)
                else:
                    printout = True
                    topDir = os.path.dirname(directory)
                    res = self.fileCatalog.listDirectory(topDir)
                    if not res["OK"]:
                        raise RuntimeError("Error listing directory: " + res["Message"])
                    matchDir = directory.split("...")[0]
                    directories += [
                        d
                        for d in res["Value"]["Successful"].get(topDir, {}).get("SubDirs", [])
                        if d.startswith(matchDir)
                    ]
            if printout:
                gLogger.notice("Expanded list of %d directories:\n%s" % (len(directories), "\n".join(directories)))
            return directories
        try:
            bkQuery = self.__getBKQuery()
        except ValueError:
            bkQuery = None
        if bkQuery and set(bkQuery.getQueryDict()) - {"Visible", "Production", "FileType"}:
            return bkQuery.getDirs()
        if self.prod:
            if bkQuery:
                fileType = bkQuery.getFileTypeList()
            else:
                fileType = []
            res = self.transClient.getTransformationParameters(self.prod, ["OutputDirectories"])
            if not res["OK"]:
                raise RuntimeError("Error getting transformation parameters: " + res["Message"])
            directories = []
            dirList = res["Value"]
            if isinstance(dirList, str) and dirList[0] == "[" and dirList[-1] == "]":
                dirList = ast.literal_eval(dirList)
            for dirName in dirList:
                # There is a shortcut when multiple streams are used, only the stream name is repeated!
                if ";" in dirName:
                    items = dirName.split(";")
                    baseDir = os.path.dirname(items[0])
                    items[0] = os.path.basename(items[0])
                    lastItems = items[-1].split("/")
                    items[-1] = lastItems[0]
                    if len(lastItems) > 1:
                        suffix = "/".join(lastItems[1:])
                    else:
                        suffix = ""
                    for it in items:
                        directories.append(os.path.join(baseDir, it, suffix))
                else:
                    if dirName.endswith("/0"):
                        dirName = dirName.replace("/0", "/%08d" % int(self.prod))
                    ftOK = bool(not fileType)
                    for ft in fileType:
                        if ft in dirName:
                            ftOK = True
                            break
                    if ftOK:
                        directories.append(dirName)
            return directories
        else:
            raise RuntimeError("No files found: you need to specify either the directories or a production ID")

        ########################################################################

    def _getBKMetadata(self, lfns):
        if self.prod == 0:
            return self._getBKMetadataForFiles(lfns)
        else:
            return self._getBKMetadataBulk(lfns)

    def _getBKMetadataForFiles(self, lfns):
        """get metadata (i.e. replica flag) of a list of LFNs"""
        missingLFNs = []
        noFlagLFNs = {}
        okLFNs = []
        chunkSize = 1000
        progressBar = ProgressBar(
            len(lfns),
            title=f"Getting {len(lfns)} files metadata from BK",
            chunk=chunkSize,
            interactive=self.interactive,
        )
        for lfnChunk in breakListIntoChunks(lfns, chunkSize):
            progressBar.loop()
            for _ in range(10):
                res = self.bkClient.getFileMetadata(lfnChunk)
                if not res["OK"]:
                    gLogger.error("\nCan't get the BK metadata, retry: ", res["Message"])
                    continue
                if res["Value"]["Failed"]:
                    gLogger.error(f"Failed to get metadata for {len(res['Value']['Failed'])} files")
                    continue
                metadata = res["Value"]["Successful"]
                missingLFNs += [lfn for lfn in lfnChunk if lfn not in metadata]
                noFlagLFNs.update(
                    {lfn: metadata[lfn]["RunNumber"] for lfn in metadata if metadata[lfn]["GotReplica"] == "No"}
                )
                okLFNs += [lfn for lfn in metadata if metadata[lfn]["GotReplica"] == "Yes"]
                break
            else:
                raise RuntimeError(f"Failed to get metadata for {len(lfnChunk)} files")
        progressBar.endLoop()
        return missingLFNs, noFlagLFNs, okLFNs

    def _getBKMetadataBulk(self, lfns):
        """get metadata (i.e. replica flag) of a list of LFNs"""
        missingLFNs = []
        noFlagLFNs = {}
        okLFNs = []

        gLogger.info("Getting metadata from the BK")
        start = time.time()
        res = returnValueOrRaise(
            self.bkClient.getFilesWithMetadata(
                {
                    "Production": self.prod,
                    "Visible": "ALL",
                    "ReplicaFlag": "ALL",
                    "OnlyParameters": ["FileName", "GotReplica", "RunNumber"],
                }
            )
        )
        gLogger.info(f"Found {len(res['Records'])} files in the BK in {time.time() - start:.1f} seconds")
        idx_lfn = res["ParameterNames"].index("FileName")
        idx_replica = res["ParameterNames"].index("GotReplica")
        idx_run = res["ParameterNames"].index("RunNumber")
        lfn_metadata = {x[idx_lfn]: {"GotReplica": x[idx_replica], "RunNumber": x[idx_run]} for x in res["Records"]}
        for lfn in lfns:
            if lfn not in lfn_metadata:
                missingLFNs.append(lfn)
            elif lfn_metadata[lfn]["GotReplica"] == "No":
                noFlagLFNs[lfn] = lfn_metadata[lfn]["RunNumber"]
            else:
                okLFNs.append(lfn)
        return missingLFNs, noFlagLFNs, okLFNs

    def checkBK2TS(self):
        """check that files present in the BK are also in the FC (re-check of BKWatchAgent)"""
        bkQuery = self.__getBKQuery(fromTS=True)
        lfnsReplicaYes = self._getBKFiles(bkQuery)
        proc, nonProc, _statuses = self._getTSFiles()
        self.filesInBKNotInTS = list(set(lfnsReplicaYes) - set(proc + nonProc))
        if self.filesInBKNotInTS:
            nOfFilesInBKNotInTS = len(self.filesInBKNotInTS)
            errMsg = f"There {'is' if nOfFilesInBKNotInTS == 1 else 'are'} {nOfFilesInBKNotInTS} "
            errMsg += f"file{'' if nOfFilesInBKNotInTS == 1 else 's'} in BK that "
            errMsg += "{'is' if nOfFilesInBKNotInTS == 1 else 'are'} not in TS: {self.filesInBKNotInTS}"
            gLogger.error(errMsg)

    def checkFC2SE(self, bkCheck=True):  # pylint: disable=arguments-differ
        self.checkFC2BK(bkCheck=bkCheck)
        if self.existLFNsBKRepYes or self.existLFNsBKRepNo:
            repDict = self.compareChecksum(self.existLFNsBKRepYes + list(self.existLFNsBKRepNo))
            self.existLFNsNoSE = repDict["MissingReplica"]
            self.existLFNsNotExisting = repDict["MissingAllReplicas"]
            self.existLFNsBadReplicas = repDict["SomeReplicasCorrupted"]
            self.existLFNsBadFiles = repDict["AllReplicasCorrupted"]
            self.notRegisteredAtSE = repDict["NotRegisteredAtSE"]

    def checkSE(self, seList):
        """
        Check if the provided files are registered in the FC in a given list of SEs
        """
        lfnsReplicaNo, lfnsReplicaYes = self.__getLFNsFromBK()
        if not lfnsReplicaNo and not lfnsReplicaYes:
            lfns, notPresent = self.__getLFNsFromFC()
        else:
            lfns = lfnsReplicaYes
            notPresent = []
        gLogger.notice(f"Checking presence of {len(lfns)} files at {', '.join(seList)}")
        replicaRes = self.dataManager.getReplicas(lfns)
        if not replicaRes["OK"]:
            gLogger.error("Error getting replicas", replicaRes["Message"])
            return
        seSet = set(seList)
        success = replicaRes["Value"]["Successful"]
        self.absentLFNsInFC = sorted(set(notPresent) | set(replicaRes["Value"]["Failed"]))
        self.existLFNsNoSE = [lfn for lfn in success if not seSet & set(success[lfn])]

    def compareChecksum(self, lfns):
        """compare the checksum of the file in the FC and the checksum of the physical replicas.
        Returns a dictionary containing 3 sub-dictionaries: one with files with missing PFN, one with
        files with all replicas corrupted, and one with files with some replicas corrupted and at least
        one good replica
        """
        retDict = {
            "AllReplicasCorrupted": {},
            "SomeReplicasCorrupted": {},
            "MissingReplica": {},
            "MissingAllReplicas": {},
            "NoReplicas": {},
            "NotRegisteredAtSE": {},
        }

        chunkSize = 100
        replicas = {}
        setLfns = set(lfns)
        cachedLfns = setLfns & set(self.cachedReplicas)
        for lfn in cachedLfns:
            replicas[lfn] = self.cachedReplicas[lfn]
        lfnsLeft = setLfns - cachedLfns
        if lfnsLeft:
            progressBar = ProgressBar(
                len(lfnsLeft),
                title=f"Get replicas for {len(lfnsLeft)} files",
                chunk=chunkSize,
                interactive=self.interactive,
            )
            for lfnChunk in breakListIntoChunks(lfnsLeft, chunkSize):
                progressBar.loop()
                replicasRes = self.dataManager.getReplicas(lfnChunk, getUrl=False)
                if not replicasRes["OK"]:
                    raise RuntimeError("Error getting replicas:  " + replicasRes["Message"])
                replicasRes = replicasRes["Value"]
                if replicasRes["Failed"]:
                    retDict["NoReplicas"].update(replicasRes["Failed"])
                replicas.update(replicasRes["Successful"])
            progressBar.endLoop()

        # Reduce the set of files to those at requested SEs if specified
        otherReplicas = []
        if self._seList:
            notAtSE = []
            for lfn, ses in replicas.items():
                if set(ses) - self._seList:
                    otherReplicas.append(lfn)
                replicas[lfn] = set(ses) & self._seList
                if not replicas[lfn]:
                    notAtSE.append(lfn)
            if notAtSE:
                gLogger.notice(
                    f"{len(notAtSE)} files are not registered at requested SEs, check if they exist in SE..."
                )
                foundInSE = self._checkFilesInSE(notAtSE, self._seList)
                if foundInSE:
                    gLogger.notice(
                        f"Of these, {len(foundInSE)} files were found at requested SEs but are not registered"
                    )
                    retDict["NotRegisteredAtSE"] = foundInSE
                else:
                    gLogger.notice("None of them were found at requested SEs, ignore them")
                for lfn in [lfn for lfn in notAtSE if lfn not in foundInSE]:
                    del replicas[lfn]

        if not replicas:
            return retDict
        progressBar = ProgressBar(
            len(replicas),
            title=f"Get FC metadata for {len(replicas)} files to be checked: ",
            chunk=chunkSize,
            interactive=self.interactive,
        )
        metadata = {}
        for lfnChunk in breakListIntoChunks(replicas, chunkSize):
            progressBar.loop()
            res = self.fileCatalog.getFileMetadata(lfnChunk)
            if not res["OK"]:
                raise RuntimeError("Error getting file metadata: " + res["Message"])
            metadata.update(res["Value"]["Successful"])
        progressBar.endLoop()

        gLogger.notice("Check existence and compare checksum file by file...")
        csDict = {}
        seFiles = {}
        # Reverse the LFN->SE dictionary
        nReps = 0
        for lfn, lfnReplicas in replicas.items():
            csDict.setdefault(lfn, {})["LFCChecksum"] = metadata.get(lfn, {}).get("Checksum")
            for se in lfnReplicas:
                seFiles.setdefault(se, []).append(lfn)
                nReps += 1

        gLogger.notice("Getting checksum of %d replicas in %d SEs" % (nReps, len(seFiles)))
        checkSum = {}
        lfnNotExisting = {}
        lfnNoInfo = {}
        logLevel = gLogger.getLevel()
        gLogger.setLevel("FATAL")
        for num, se in enumerate(sorted(seFiles)):
            progressBar = ProgressBar(
                len(seFiles[se]),
                title="%d. At %s (%d files)" % (num, se, len(seFiles[se])),
                chunk=chunkSize,
                interactive=self.interactive,
            )
            oSe = StorageElement(se)
            errMsg = None
            notFound = 0
            for surlChunk in breakListIntoChunks(seFiles[se], chunkSize):
                progressBar.loop()
                metadata = oSe.getFileMetadata(surlChunk)
                if not metadata["OK"]:
                    errMsg = f"Error: getFileMetadata returns {metadata['Message']}. Ignore those replicas"
                    progressBar.comment(errMsg)
                    # Remove from list of replicas as we don't know whether it is OK or not
                    for lfn in seFiles[se]:
                        lfnNoInfo.setdefault(lfn, []).append(se)
                else:
                    metadata = metadata["Value"]
                    # print metadata
                    notFound += len(metadata["Failed"])
                    for lfn in metadata["Failed"]:
                        lfnNotExisting.setdefault(lfn, []).append(se)
                    for lfn in metadata["Successful"]:
                        checkSum.setdefault(lfn, {})[se] = metadata["Successful"][lfn]["Checksum"]
            if notFound:
                errMsg = "%d files not found" % notFound
            progressBar.endLoop(errMsg)
        gLogger.setLevel(logLevel)

        gLogger.notice(f"Verifying checksum of {len(replicas)} files")
        for lfn, replicaDict in replicas.items():
            # get the lfn checksum from the FC
            oneGoodReplica = bool(lfn in otherReplicas)
            allGoodReplicas = True
            lfcChecksum = csDict[lfn].pop("LFCChecksum")
            for se in replicaDict:
                # If replica doesn't exist skip check
                if se in lfnNotExisting.get(lfn, []):
                    allGoodReplicas = False
                    continue
                if se in lfnNoInfo.get(lfn, []):
                    # If there is no info, a priori it could be good
                    oneGoodReplica = True
                    continue
                # get the surls metadata and compare the checksum
                surlChecksum = checkSum.get(lfn, {}).get(se, "")
                if not surlChecksum or not compareAdler(lfcChecksum, surlChecksum):
                    # if lfcChecksum does not match surlChecksum
                    csDict[lfn][se] = {"PFNChecksum": surlChecksum}
                    gLogger.info(
                        "ERROR!! checksum mismatch at %s for LFN %s:  LFC checksum: %s , PFN checksum : %s "
                        % (se, lfn, lfcChecksum, surlChecksum)
                    )
                    allGoodReplicas = False
                else:
                    oneGoodReplica = True
            if not oneGoodReplica:
                if lfn in lfnNotExisting:
                    gLogger.info("=> All replicas are missing", lfn)
                    retDict["MissingAllReplicas"][lfn] = "All"
                else:
                    gLogger.info("=> All replicas have bad checksum", lfn)
                    retDict["AllReplicasCorrupted"][lfn] = csDict[lfn]
            elif not allGoodReplicas:
                if lfn in lfnNotExisting:
                    gLogger.info("=> At least one replica missing", lfn)
                    retDict["MissingReplica"][lfn] = lfnNotExisting[lfn]
                else:
                    gLogger.info("=> At least one replica with good Checksum", lfn)
                    retDict["SomeReplicasCorrupted"][lfn] = csDict[lfn]

        return retDict

    ################################################################################
    # properties

    def set_prod(self, value):
        """Setter"""
        if value:
            value = int(value)
            res = self.transClient.getTransformation(value, extraParams=False)
            if not res["OK"]:
                raise RuntimeError("Couldn't find transformation %d: %s" % (value, res["Message"]))
            self.transType = res["Value"]["Type"]
            if self.interactive:
                gLogger.info("Production %d has type %s" % (value, self.transType))
        else:
            value = 0
        self._prod = value
        self._daughtersCache = None

    def get_prod(self):
        """Getter"""
        return self._prod

    prod = property(get_prod, set_prod)

    def set_fileType(self, value):
        """Setter"""
        fts = [ft.upper() for ft in value]
        self._fileType = fts

    def get_fileType(self):
        """Getter"""
        return self._fileType

    fileType = property(get_fileType, set_fileType)

    def set_fileTypesExcluded(self, value):
        """Setter"""
        fts = [ft.upper() for ft in value]
        self._fileTypesExcluded = fts

    def get_fileTypesExcluded(self):
        """Getter"""
        return self._fileTypesExcluded

    fileTypesExcluded = property(get_fileTypesExcluded, set_fileTypesExcluded)

    def set_bkQuery(self, value):
        """Setter"""
        if isinstance(value, str):
            self._bkQuery = ast.literal_eval(value)
        else:
            self._bkQuery = value

    def get_bkQuery(self):
        """Getter"""
        return self._bkQuery

    bkQuery = property(get_bkQuery, set_bkQuery)

    def set_lfns(self, value):
        """Setter"""
        if isinstance(value, str):
            value = [value]
        value = [v.replace(" ", "").replace("//", "/") for v in value]
        self._lfns = value

    def get_lfns(self):
        """Getter"""
        return self._lfns

    lfns = property(get_lfns, set_lfns)

    def set_status(self, value):
        """Setter"""
        self._status = value

    def get_status(self):
        """Getter"""
        return self._status

    status = property(get_status, set_status)

    def set_verbose(self, value):
        """Setter"""
        self._verbose = bool(value)

    def get_verbose(self):
        """Getter"""
        return self._verbose

    verbose = property(get_verbose, set_verbose)

    def set_seList(self, value):
        """Setter"""
        self._seList = set(resolveSEGroup(value))

    def get_seList(self):
        """Getter"""
        return self._seList

    seList = property(get_seList, set_seList)

    def _findNextProduction(self):
        """
        Find in the next productions one that uses the current production as input in the BK query
        Returns its number and its type
        """
        for nextProd in range(self.prod + 1, self.prod + 6):
            res = self.transClient.getBookkeepingQuery(nextProd)
            if res["OK"] and res["Value"].get("ProductionID") == self.prod:
                res = self.transClient.getTransformation(nextProd)
                if not res["OK"]:
                    return None, None
                return (nextProd, res["Value"]["Type"])
        return None, None



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the Start or End Run for a given Transformation of add a set runs.

Examples of Usage :
  with 1234 = ProdID
  with 99000 = RunNumber
  dirac-production-set-run 1234 --List        (list of runs.)
  dirac-production-set-run 1234 --AddRuns 99000   (only for list of runs.)
  dirac-production-set-run 1234 --EndRun 99000   (change endrun.)
  dirac-production-set-run 1234 --StartRun 99000 (change startrun.)
"""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from DIRAC import gLogger

    Script.setUsageMessage(
        f"{__doc__}"
        f"Usage:\n"
        f"  {Script.scriptName} [option|cfgfile] ... Prod -Option [RunNumber|RunList]\n"
        "Arguments:\n"
        "  Prod:      DIRAC Production Id\n"
        "  RunNumber: New Start or End run\n"
        "  RunList: List of Runs to be added\n"
        "Examples:\n"
        "  # show the list of runs for transformation 92\n"
        "  dirac-production-set-run.py 92 --List\n"
        "  # add some discrete run to transformation 92\n"
        "  dirac-production-set-run.py 92 --AddRuns 98200,98201\n"
        "  # add some discrete run and a range of runs to transformation 92\n"
        "  dirac-production-set-run.py 92 --AddRuns 98200,98201,99000:99100\n"
        "  # change the start run for transformation 92\n"
        "  dirac-production-set-run.py 92 --StartRun 98200\n"
        "  # change the end run  for transformation 92\n"
        "  dirac-production-set-run.py 92 --EndRun   98200\n"
    )

    Script.registerSwitch("", "EndRun=", "Specify endrun for the transformation")
    Script.registerSwitch("", "StartRun=", "Specify startrun for the transformation")
    Script.registerSwitch("", "AddRuns=", "add List of runs to the transformation")
    Script.registerSwitch("", "List", "List the runs for the transformation")

    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

    transList = getTransformations(Script.getPositionalArgs())

    settings = {}
    for opt, val in Script.getUnprocessedSwitches():
        if opt in ("StartRun", "EndRun"):
            try:
                settings[opt] = int(val)
            except TypeError:
                gLogger.error("Invalid run number:", str(val))
                DIRAC.exit(1)
        elif opt == "AddRuns":
            try:
                settings[opt] = [int(runID) for runID in val.split(",")]
            except TypeError:
                gLogger.error("Invalid run list", str(val))
                DIRAC.exit(1)
        elif opt == "List":
            settings[opt] = True
    if "AddRuns" in settings and ("StartRun" in settings or "EndRun" in settings):
        gLogger.error("Incompatible requests, cannot set run list and start/end run")
        DIRAC.exit(1)

    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    client = TransformationClient()

    for prodId in transList:
        res = client.getBookkeepingQuery(prodId)
        if not res["OK"]:
            gLogger.error(f"Error retrieving BKQuery for transformation {prodId}", res["Message"])
            DIRAC.exit(2)
        bkDict = res["Value"]
        startRun = bkDict.get("StartRun", 0)
        endRun = bkDict.get("EndRun", 0)
        runNumbers = bkDict.get("RunNumbers", "All")

        if ("StartRun" in settings or "EndRun" in settings) and runNumbers and runNumbers != "All":
            gLogger.notice("Transformation %d has RunNumbers key" % prodId)
            settings = {"List": True}

        if "AddRuns" in settings and (startRun or endRun):
            gLogger.notice("Transformation %d has start run or end run: %s:%s" % (prodId, str(startRun), str(endRun)))
            settings = {"List": True}

        if "AddRuns" in settings and (not runNumbers or runNumbers == "All"):
            gLogger.notice("Transformation %d doesn't have RunNumbers key or set to All" % prodId)
            settings = {"List": True}

        if not isinstance(runNumbers, list):
            runNumbers = [runNumbers]

        changed = False
        if "StartRun" in settings:
            changed = True
            runId = settings["StartRun"]
            res = client.setBookkeepingQueryStartRun(prodId, runId)
            if res["OK"]:
                gLogger.notice("Start run of transformation %d is now %d" % (prodId, runId))
                startRun = runId
            else:
                gLogger.error("Error setting start run", res["Message"])

        if "EndRun" in settings:
            changed = True
            runId = settings["EndRun"]
            res = client.setBookkeepingQueryEndRun(prodId, runId)
            if res["OK"]:
                gLogger.notice("End run of transformation %d is now %d" % (prodId, runId))
                endRun = runId
            else:
                gLogger.error("Error setting end run", res["Message"])

        if "AddRuns" in settings:
            changed = True
            runList = [int(run) for run in settings["AddRuns"] if run not in runNumbers]
            res = client.addBookkeepingQueryRunList(prodId, runList)
            if res["OK"]:
                gLogger.notice("Run list modified for transformation %d" % prodId)
                runNumbers += runList
            else:
                gLogger.error("Error modifying run list:", res["Message"])

        if "List" in settings:
            gLogger.notice("%sRun selection settings for transformation %d:" % ("\n" if changed else "", prodId))
            if runNumbers != ["All"]:
                gLogger.notice(f"List of runs: {','.join([str(run) for run in sorted(runNumbers)])}")
            else:
                if startRun:
                    gLogger.notice(f"Start run is {startRun}")
                else:
                    gLogger.notice("No start run defined ")
                if endRun:
                    gLogger.notice(f"End run is {endRun}")
                else:
                    gLogger.notice("No end run defined ")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Gaudi Application module - main module: creates the environment,
    executes gaudirun with the right options

    This is the module used for each and every job of productions. It can also be used by users.
"""
import json
import os

from DIRAC import S_OK, S_ERROR, gLogger, gConfig

from LHCbDIRAC.Core.Utilities.ProductionOptions import getDataOptions, getModuleOptions
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.RunApplication import RunApplication, LHCbApplicationError, LHCbDIRACError


class GaudiApplication(ModuleBase):
    """GaudiApplication class."""

    def __init__(self, bkClient=None, dm=None):
        """Usual init for LHCb workflow modules."""

        self.log = gLogger.getSubLogger("GaudiApplication")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.systemConfig = ""
        self.stdError = ""
        self.inputDataType = "MDF"
        self.stepInputData = []  # to be resolved
        self.poolXMLCatName = "pool_xml_catalog.xml"
        self.optionsFile = ""
        self.optionsLine = ""
        self.extraOptionsLine = ""
        self.eventTimeout = None
        self.extraPackages = ""
        self.jobType = ""

    def _resolveInputVariables(self):
        """Resolve all input variables for the module here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_id=None,
        step_number=None,
    ):
        """The main execution method of GaudiApplication.

        It runs a gaudirun app using RunApplication module. This is the
        module used for each and every job of productions. It can also be
        used by users.
        """

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            if not self._checkWFAndStepStatus():
                return S_OK()

            self._resolveInputVariables()

            self.log.info(
                "Executing application %s %s for binary tag %s"
                % (self.applicationName, self.applicationVersion, self.systemConfig)
            )
            if self.jobType.lower() == "merge" in self.siteName:
                self._disableWatchdogCPUCheck()

            # Resolve options files
            if self.optionsFile.startswith("{"):
                commandOptions = json.loads(self.optionsFile)
                self.log.info("Found lbexec style configuration:", commandOptions)
            else:
                commandOptions = []
                if self.optionsFile and self.optionsFile != "None":
                    commandOptions += self.optionsFile.split(";")
                self.log.info("Final options files:", ", ".join(commandOptions))

            runNumberGauss = 0
            firstEventNumberGauss = 0
            if self.applicationName.lower() == "gauss" and self.production_id and self.prod_job_id:
                if self.jobType.lower() == "user":
                    eventsMax = self.numberOfEvents
                else:
                    # maintaining backward compatibility
                    eventsMax = self.maxNumberOfEvents if self.maxNumberOfEvents else self.numberOfEvents
                runNumberGauss = int(self.production_id) * 100 + int(self.prod_job_id)
                firstEventNumberGauss = eventsMax * (int(self.prod_job_id) - 1) + 1

            if self.optionsLine or self.jobType.lower() == "user":
                self.log.debug("Won't get any step outputs (USER job)")
                stepOutputs = []
                stepOutputTypes = []
                histogram = False
            else:
                self.log.debug("Getting the step outputs")
                stepOutputs, stepOutputTypes, histogram = self._determineOutputs()
                self.log.debug(
                    "stepOutputs, stepOutputTypes, histogram  ==>  %s, %s, %s"
                    % (stepOutputs, stepOutputTypes, histogram)
                )

            # Simple check for slow processors: auto increase of Event Timeout
            eventTimeout = self.eventTimeout
            cpuNormalization = int(gConfig.getValue("/LocalSite/CPUNormalizationFactor", 10))
            if cpuNormalization < 10:
                if not eventTimeout:
                    eventTimeout = 3600
                eventTimeout = int(eventTimeout * 10 / cpuNormalization)

            if self.optionsLine or self.jobType.lower() == "user":
                # Prepare standard project run time options
                generatedOpts = "gaudi_extra_options.py"
                if os.path.exists(generatedOpts):
                    os.remove(generatedOpts)
                inputDataOpts = getDataOptions(
                    self.applicationName, self.stepInputData, self.inputDataType, self.poolXMLCatName
                )[
                    "Value"
                ]  # always OK
                projectOpts = getModuleOptions(
                    self.applicationName,
                    self.numberOfEvents,
                    inputDataOpts,
                    self.optionsLine,
                    runNumberGauss,
                    firstEventNumberGauss,
                    self.jobType,
                )[
                    "Value"
                ]  # always OK
                self.log.info(f"Extra options generated for {self.applicationName} {self.applicationVersion} step:")
                print(
                    projectOpts
                )  # Always useful to see in the logs (don't use gLogger as we often want to cut n' paste)
                with open(generatedOpts, "w") as options:
                    options.write(projectOpts)
                # TODO Don't do this with lbexec
                commandOptions.append(generatedOpts)

            # How to run the application
            ra = RunApplication(
                self,
                commandOptions,
                stepOutputTypes,
                histogram,
                runNumberGauss,
                firstEventNumberGauss,
                eventTimeout,
            )

            # Now really running
            self.setApplicationStatus(f"{self.applicationName} step {self.step_number}")
            ra.run()  # This would trigger an exception in case of failure, or application status != 0

            self.log.info(f"Going to manage {self.applicationName} output")
            self._manageAppOutput(stepOutputs)

            # Still have to set the application status e.g. user job case.
            self.setApplicationStatus(f"{self.applicationName} {self.applicationVersion} Successful")

            return S_OK(f"{self.applicationName} {self.applicationVersion} Successful")
        except LHCbApplicationError as lbae:  # This is the case for real application errors
            self.setApplicationStatus(repr(lbae))
            return S_ERROR(str(lbae))
        except LHCbDIRACError as lbde:  # This is the case for LHCbDIRAC errors (e.g. subProcess call failed)
            self.setApplicationStatus(repr(lbde))
            return S_ERROR(str(lbde))
        except Exception as exc:  # pylint:disable=broad-except
            self.log.exception("Failure in GaudiApplication execute module", lException=exc, lExcInfo=True)
            self.setApplicationStatus("Error in GaudiApplication module")
            return S_ERROR(str(exc))
        finally:
            super().finalize()
